open HolKernel Parse boolLib bossLib;
open finite_mapTheory;
open factoredSystemTheory;
open sublistTheory;
open listTheory;
open pred_setTheory;
open fmap_utilsTheory
open list_utilsTheory

val _ = new_theory "actionSeqProcess";

val sat_precond_as_def = 
    Define `(sat_precond_as(s, a::as) = (FST a SUBMAP s /\ sat_precond_as(state_succ s a, as))) 
    /\ (sat_precond_as(s, []) = T)`; 

val sat_precond_as_pair = store_thm("sat_precond_as_pair",
``(sat_precond_as(s, (p,e)::as) = (p SUBMAP s /\ sat_precond_as(state_succ s (p,e), as))) 
    /\ (sat_precond_as(s, []) = T)``,
METIS_TAC[sat_precond_as_def, pairTheory.PAIR, pairTheory.FST, pairTheory.SND])

val rem_effectless_act_def 
   = Define `(rem_effectless_act(a::as)
	         = if (FDOM(SND a) <> EMPTY) then a::rem_effectless_act(as)
	     	   else rem_effectless_act(as))
	   /\ (rem_effectless_act([]) = [])`;

val no_effectless_act_def
    = Define `(no_effectless_act(a::as)
		= ((FDOM (SND a) <> EMPTY) /\ no_effectless_act(as)))
	    /\ (no_effectless_act([]) = T)` ;

val graph_plan_lemma_4 = store_thm("graph_plan_lemma_4",
``!s s' as vs P. ((!a. (MEM a as /\ P a) ==> DISJOINT (FDOM (SND a)) vs)
	    /\ sat_precond_as(s, as)
	    /\ sat_precond_as(s', FILTER (\a. ~(P a)) as)
	    /\ (DRESTRICT s vs = DRESTRICT s' vs)) 
     	    ==>
            ((DRESTRICT (exec_plan(s,as)) (vs)) = (DRESTRICT (exec_plan (s',FILTER (\a. ~(P a)) as)) (vs)))``,
Induct_on`as`
THEN SRW_TAC[][exec_plan_def]
THENL
[
	FULL_SIMP_TAC(srw_ss())[sat_precond_as_def] 	
	THEN `DRESTRICT (state_succ s h) vs = DRESTRICT (state_succ s' h) vs` by SRW_TAC[][graph_plan_lemma_4_1]
	THEN METIS_TAC[]
	,
	FULL_SIMP_TAC(srw_ss())[sat_precond_as_def] 	
	THEN `DRESTRICT (state_succ s h) (vs) = DRESTRICT s (vs)` by METIS_TAC[GSYM graph_plan_lemma_1_2]
	THEN METIS_TAC[]
]);

val rem_condless_act_def 
   = Define `(rem_condless_act(s, pfx_a, a::as) = if ((FST a) SUBMAP exec_plan(s, pfx_a)) then rem_condless_act(s, pfx_a++[a], as)
	     else rem_condless_act(s, pfx_a, as))
	   /\ (rem_condless_act(s, pfx_a, [] ) = pfx_a)`;

val rem_condless_act_pair = store_thm("rem_condless_act_pair",
``(rem_condless_act(s, pfx_a, (p,e)::as) = if (p SUBMAP exec_plan(s, pfx_a)) then rem_condless_act(s, pfx_a++[(p,e)], as)
	     else rem_condless_act(s, pfx_a, as))
	   /\ (rem_condless_act(s, pfx_a, [] ) = pfx_a)``,
METIS_TAC[rem_condless_act_def, pairTheory.PAIR, pairTheory.FST, pairTheory.SND])

val graph_plan_lemma_7_1_1 = store_thm("graph_plan_lemma_7_1_1",
``!s h as pfx. exec_plan (s,rem_condless_act (s,h:: pfx,as)) 
       	       		 = exec_plan (state_succ s h,rem_condless_act (state_succ s h ,pfx,as))``,
Induct_on`as`
THEN SRW_TAC[][rem_condless_act_def, exec_plan_def, state_succ_def] 
);

val graph_plan_lemma_7_1 = store_thm("graph_plan_lemma_7_1",
``!as s. (exec_plan(s, as) = exec_plan(s, rem_condless_act(s, [], as)))``,
Induct_on`as`
THEN SRW_TAC[][rem_condless_act_def, exec_plan_def]
THEN METIS_TAC[graph_plan_lemma_7_1_1, FDOM_state_succ, state_succ_def]
);

val graph_plan_lemma_7_2_1_1 = store_thm("graph_plan_lemma_7_2_1_1",
``!h' pfx as s. rem_condless_act (s,h'::pfx,as) = h'::rem_condless_act (state_succ s h',pfx,as)``,
Induct_on`as`
THEN SRW_TAC[][rem_condless_act_def, exec_plan_def, state_succ_def] 
);

val graph_plan_lemma_7_2_1 = store_thm("graph_plan_lemma_7_2_1",
``! h h' as as' s. h'::as' <<= rem_condless_act (s,[h],as) ==> (as' <<=  rem_condless_act (state_succ s h,[],as) /\ (h' = h))``,
Induct_on`as`
THEN SRW_TAC[][rem_condless_act_def, exec_plan_def]
THEN FULL_SIMP_TAC(srw_ss())[graph_plan_lemma_7_2_1_1]);


val graph_plan_lemma_7_2 = store_thm("graph_plan_lemma_7_2",
``!as s.  sat_precond_as(s, rem_condless_act (s,[],as))``,
Induct_on`as`
THEN SRW_TAC[][rem_condless_act_def, exec_plan_def, sat_precond_as_def, graph_plan_lemma_7_2_1_1]);

val graph_plan_lemma_7_3 = store_thm("graph_plan_lemma_7_3",
``!as s. (LENGTH (rem_condless_act(s, [], as)) <= LENGTH as)``,
Induct_on`as`
THEN SRW_TAC[][rem_condless_act_def, exec_plan_def]
THENL
[
	FULL_SIMP_TAC(srw_ss())[graph_plan_lemma_7_2_1_1]
	THEN SRW_TAC[][graph_plan_lemma_7_2_1_1]
	,
	`LENGTH (rem_condless_act (s,[],as)) ≤ (LENGTH as)` by METIS_TAC[]
	THEN DECIDE_TAC
]);

val graph_plan_lemma_7_4 = store_thm("graph_plan_lemma_7_4",
``!as A s. (set as SUBSET A) ==>
	  (set (rem_condless_act (s,[],as)) SUBSET A )``,
Induct_on`as`
THEN SRW_TAC[][rem_condless_act_def, exec_plan_def]
THEN FULL_SIMP_TAC(srw_ss())[graph_plan_lemma_7_2_1_1]
THEN METIS_TAC[graph_plan_lemma_7_2_1_1, FDOM_state_succ]);

val graph_plan_lemma_7_6 = store_thm("graph_plan_lemma_7_6",
``!as s P. (LENGTH (FILTER P (rem_condless_act (s,[],as)))  <= LENGTH (FILTER P as) )``,
Induct_on`as`
THEN SRW_TAC[][rem_condless_act_def, exec_plan_def]
THENL
[
	FULL_SIMP_TAC(srw_ss())[graph_plan_lemma_7_2_1_1]
	THEN METIS_TAC[graph_plan_lemma_7_2_1_1, FDOM_state_succ]
	,
	FULL_SIMP_TAC(srw_ss())[graph_plan_lemma_7_2_1_1]
	THEN METIS_TAC[graph_plan_lemma_7_2_1_1, FDOM_state_succ]
	,
	`LENGTH (FILTER P (rem_condless_act (s,[],as))) <= (LENGTH (FILTER P as))` by METIS_TAC[]
	THEN DECIDE_TAC
]);

val graph_plan_lemma_7_7 = store_thm("graph_plan_lemma_7_7",
``!s P as as2. (EVERY P as /\ EVERY P as2)==> EVERY (P) (rem_condless_act (s,as2,as))``,
Induct_on `as`
THEN SRW_TAC[][rem_condless_act_def]);

val graph_plan_lemma_7_8 = store_thm("graph_plan_lemma_7_8",
``!s as. sublist (rem_condless_act(s, [], as)) as``,
Induct_on `as`
THEN SRW_TAC[SatisfySimps.SATISFY_ss][rem_condless_act_def, graph_plan_lemma_7_8_1]
THEN METIS_TAC[graph_plan_lemma_7_2_1_1, sublist_def]);

val graph_plan_lemma_7_10 = store_thm("graph_plan_lemma_7_10",
``!PROB as. as IN (valid_plans PROB) ==> rem_condless_act(s, [], as) IN (valid_plans PROB)``,
SRW_TAC[][valid_plans_def]
THEN METIS_TAC[graph_plan_lemma_7_1, graph_plan_lemma_7_4]);

val graph_plan_lemma_7 = store_thm("graph_plan_lemma_7",
``!as A s. (exec_plan(s, as) = exec_plan(s, rem_condless_act(s, [], as)))
	  /\ sat_precond_as(s, rem_condless_act(s, [], as))
	  /\ (LENGTH (rem_condless_act(s, [], as)) <= LENGTH as) 
	  /\ ((set as SUBSET A) ==> (set (rem_condless_act (s,[],as)) SUBSET A ))
	  /\ (!P. (LENGTH (FILTER P (rem_condless_act (s,[],as)))  <= LENGTH (FILTER P as) ))``,
METIS_TAC[graph_plan_lemma_7_1, graph_plan_lemma_7_2, graph_plan_lemma_7_3, graph_plan_lemma_7_4, graph_plan_lemma_7_6]);

val parent_children_lemma_1_1_1_10_1 = store_thm("parent_children_lemma_1_1_1_10_1",
``!s1 s2 as. (s1 SUBMAP s2) /\ sat_precond_as(s1, as )
                 ==> sat_precond_as(s2, as)``,
Induct_on `as`
THEN SRW_TAC[][sat_precond_as_def]
THEN1 METIS_TAC[SUBMAP_TRANS]
THEN FIRST_X_ASSUM MATCH_MP_TAC
THEN Q.EXISTS_TAC `state_succ s1 h`
THEN SRW_TAC[][]
THEN METIS_TAC[parent_children_lemma_1_1_1_10_1_1])

val submap_init_submap_exec = store_thm("submap_init_submap_exec",
``!s1 s2. s1 SUBMAP s2 /\ sat_precond_as(s1, as ) ==>
  exec_plan(s1, as) SUBMAP (exec_plan(s2, as))``,
Induct_on `as`
THEN SRW_TAC[][exec_plan_def]
THEN FIRST_X_ASSUM MATCH_MP_TAC
THEN SRW_TAC[][]
THEN FULL_SIMP_TAC(srw_ss())[sat_precond_as_def]
THEN METIS_TAC[parent_children_lemma_1_1_1_10_1_1])

val parent_children_lemma_1_1_1_10 = store_thm("parent_children_lemma_1_1_1_10",
``!vs s as. sat_precond_as(DRESTRICT s vs, as )
                 ==> sat_precond_as(s, as)``,
METIS_TAC[DRESTRICT_SUBMAP, parent_children_lemma_1_1_1_10_1])

val varset_action_def = Define ` varset_action(a, varset) <=>  ((FDOM (SND a)) SUBSET varset)`;

val varset_action_pair = store_thm("varset_action_pair",
`` varset_action((p,e), vs) <=>  ((FDOM e) SUBSET vs)``,
METIS_TAC[varset_action_def, pairTheory.PAIR, pairTheory.SND])

val graph_plan_lemma_14_2 = store_thm("graph_plan_lemma_14_2",
``!x y. (SND (x) = SND y)
     ==> ((\a. varset_action(a, vs)) x = (\a. varset_action(a, vs)) y)``,
SRW_TAC[][varset_action_def]);

val graph_plan_lemma_16_1 = store_thm("graph_plan_lemma_16_1",
``!s as. (exec_plan (s,as) = exec_plan (s,rem_effectless_act (as)))``,
Induct_on `as` 
THEN SRW_TAC[][rem_effectless_act_def, exec_plan_def, graph_plan_lemma_16_1_1]
THEN SRW_TAC[][rem_effectless_act_def, exec_plan_def, graph_plan_lemma_16_1_1]);

val graph_plan_lemma_16_2 = store_thm("graph_plan_lemma_16_2",
``!as s. ( sat_precond_as (s, as) ==> sat_precond_as (s,rem_effectless_act (as)))``,
Induct_on`as`
THEN SRW_TAC[][sat_precond_as_def, rem_effectless_act_def]
THEN FULL_SIMP_TAC(srw_ss())[graph_plan_lemma_16_1_1]);

val graph_plan_lemma_16_3 = store_thm("graph_plan_lemma_16_3",
``!as. LENGTH (rem_effectless_act (as)) ≤ LENGTH as``,
Induct_on `as`
THEN SRW_TAC[][rem_effectless_act_def]
THEN ` LENGTH (rem_effectless_act as) ≤ (LENGTH as)` by SRW_TAC[][]
THEN DECIDE_TAC);

val graph_plan_lemma_16_4 = store_thm("graph_plan_lemma_16_4",
``! A as. (set as ⊆ A ⇒ set (rem_effectless_act (as)) ⊆ A)``,
Induct_on`as`
THEN SRW_TAC[][rem_effectless_act_def]);

val graph_plan_lemma_16_4' = store_thm("graph_plan_lemma_16_4'",
``! A as. (as IN (valid_plans A)) ⇒ (rem_effectless_act (as)) IN (valid_plans A)``,
Induct_on`as`
THEN FULL_SIMP_TAC(srw_ss())[rem_effectless_act_def, valid_plans_def]
THEN SRW_TAC[][]);

val graph_plan_lemma_16_5 = store_thm("graph_plan_lemma_16_5",
``∀P as. LENGTH (FILTER P (rem_effectless_act (as))) ≤
       LENGTH (FILTER P as)``,
Induct_on`as`
THEN SRW_TAC[][rem_effectless_act_def]
THEN `LENGTH (FILTER P (rem_effectless_act as)) ≤  (LENGTH (FILTER P as))` by SRW_TAC[][]
THEN DECIDE_TAC);

val graph_plan_lemma_16_6 = store_thm("graph_plan_lemma_16_6", 
``!as. no_effectless_act (rem_effectless_act (as))``,
Induct_on`as`
THEN SRW_TAC[][no_effectless_act_def, rem_effectless_act_def]);

val graph_plan_lemma_16_7 = store_thm("graph_plan_lemma_16_7",
``!as. no_effectless_act(as) =  EVERY (\a. FDOM(SND a) <> EMPTY) as ``,
Induct_on `as`
THEN SRW_TAC[][no_effectless_act_def, EVERY_DEF]);

val graph_plan_lemma_16_8 = store_thm("graph_plan_lemma_16_8",
``!P as. (EVERY P as)==> EVERY (P) (rem_effectless_act (as))``,
Induct_on `as`
THEN SRW_TAC[][rem_effectless_act_def]);

val graph_plan_lemma_16_9 = store_thm("graph_plan_lemma_16_9",
``!as. sublist (rem_effectless_act(as)) as``,
Induct_on `as`
THEN SRW_TAC[SatisfySimps.SATISFY_ss][rem_effectless_act_def, graph_plan_lemma_7_8_1, sublist_def]);

val graph_plan_lemma_16_10 = store_thm("graph_plan_lemma_16_10",
``!as P. no_effectless_act(as) 
        ==> no_effectless_act(FILTER P as)``,
METIS_TAC[graph_plan_lemma_16_7,EVERY_FILTER_IMP]);

val graph_plan_lemma_16_11 = store_thm("graph_plan_lemma_16_11",
``!as1 as2. sublist as1 (rem_effectless_act(as2))
       ==> sublist as1 as2 ``,
SRW_TAC[][]
THEN METIS_TAC[graph_plan_lemma_16_9, sublist_trans]);

val graph_plan_lemma_16_12 = store_thm("graph_plan_lemma_16_12",
``!as1 as2. no_effectless_act(as1 ++ as2) 
       	    <=> no_effectless_act(as1) /\ no_effectless_act(as2)``,
Induct_on `as1`
THEN SRW_TAC[][no_effectless_act_def]
THEN METIS_TAC[]);


val graph_plan_lemma_16_13 = store_thm("graph_plan_lemma_16_13",
``!as1 as2. sublist as1 as2 /\ no_effectless_act(as2)
       ==>   no_effectless_act(as1)``,
Induct_on `as1`
THEN SRW_TAC[][no_effectless_act_def]
THEN MP_TAC(sublist_MEM |> INST_TYPE[alpha |-> ``:α # (β |-> γ)``] |> Q.SPECL[`h`,`as1`,`as2`])
THEN SRW_TAC[][]
THEN1 METIS_TAC[EVERY_MEM |> Q.ISPEC `(\a. FDOM(SND a) <> EMPTY)`, graph_plan_lemma_16_7]
THEN1 METIS_TAC[sublist_cons_3]);

val graph_plan_lemma_16_14 = store_thm("graph_plan_lemma_16_14",
``!PROB as. exec_plan(s, as) = exec_plan(s, rem_effectless_act(as))``,
METIS_TAC[graph_plan_lemma_16_1]
THEN METIS_TAC[graph_plan_lemma_16_4]);

val graph_plan_lemma_16 = store_thm("graph_plan_lemma_16", 
`` !s A as.   (exec_plan (s,as) = exec_plan (s,rem_effectless_act (as))) ∧
     (sat_precond_as (s, as) ==> sat_precond_as (s,rem_effectless_act (as))) ∧
     LENGTH (rem_effectless_act (as)) ≤ LENGTH as ∧
     (set as ⊆ A ⇒ set (rem_effectless_act (as)) ⊆ A) ∧
     no_effectless_act (rem_effectless_act (as)) /\
     ∀P.
       LENGTH (FILTER P (rem_effectless_act (as))) ≤
       LENGTH (FILTER P as)``,
METIS_TAC[graph_plan_lemma_16_1, graph_plan_lemma_16_2, graph_plan_lemma_16_3, graph_plan_lemma_16_4, graph_plan_lemma_16_5, graph_plan_lemma_16_6]);

val rem_effectless_act_set_def = Define`rem_effectless_act_set A = {a | a IN A /\ (FDOM (SND a) <> EMPTY)}`;

val rem_effectless_act_subset_rem_effectless_act_set_thm = store_thm("rem_effectless_act_subset_rem_effectless_act_set_thm",
``!as A. set as SUBSET A ==> (set (rem_effectless_act as)) SUBSET (rem_effectless_act_set A)``,
Induct_on `as`
THEN SRW_TAC[][rem_effectless_act_def]
THEN SRW_TAC[][rem_effectless_act_set_def])

val rem_effectless_act_set_no_empty_actions_thm = store_thm("rem_effectless_act_set_no_empty_actions_thm",
``!A. rem_effectless_act_set A SUBSET {a |FDOM (SND a) ≠ ∅}``,
SRW_TAC[][rem_effectless_act_set_def, SUBSET_DEF])

val graph_plan_lemma_7_9 = store_thm("graph_plan_lemma_7_9",
``!s as. no_effectless_act(as) ==> no_effectless_act(rem_condless_act(s, [], as))``,
Induct_on `as`
THEN SRW_TAC[SatisfySimps.SATISFY_ss][rem_condless_act_def, no_effectless_act_def]
THEN METIS_TAC[graph_plan_lemma_7_2_1_1, no_effectless_act_def]);

val graph_plan_lemma_17 = store_thm("graph_plan_lemma_17",
``!as_1 as_2 as s. (as_1++as_2 = as) /\ sat_precond_as(s,as)
==>(sat_precond_as(s,as_1) /\ sat_precond_as( exec_plan(s, as_1),as_2))``,
Induct_on`as`
THEN SRW_TAC[][exec_plan_def]
THEN Cases_on`as_1`
THEN FULL_SIMP_TAC(srw_ss())[rich_listTheory.IS_PREFIX_NIL, isPREFIX_THM, exec_plan_def, sat_precond_as_def]
THEN SRW_TAC[][]);

val vtx_cut_main_lemma_1_1_1_1 = store_thm("vtx_cut_main_lemma_1_1_1_1",
``!as. no_effectless_act as /\ (!x. ~(FDOM (SND (f x)) = EMPTY))
       ==> EVERY (λa. ~(f a = (FEMPTY,FEMPTY))) as``,
Induct_on `as` THEN SRW_TAC[][EVERY_DEF]
THEN  METIS_TAC[vtx_cut_main_lemma_1_1_1_1_1,
                graph_plan_lemma_16_7,
                no_effectless_act_def])

val vtx_cut_main_lemma_xx_7 = store_thm("vtx_cut_main_lemma_xx_7",
``!s as as'. sat_precond_as(s, as++as') ==> sat_precond_as(exec_plan(s, as) , as')``,
Induct_on `as`
THEN SRW_TAC[][sat_precond_as_def, exec_plan_def])

val child_parent_lemma_1_1_5 = store_thm("child_parent_lemma_1_1_5",
``!PROB a vs. ((a IN PROB) /\ ( ~varset_action (a,vs)))
    ==> ~(DISJOINT (FDOM (SND a)) ((prob_dom PROB) DIFF vs))``,
SRW_TAC[][]
THEN MP_TAC(FDOM_eff_subset_prob_dom_pair |> Q.SPEC `a`)
THEN SRW_TAC[][]
THEN FULL_SIMP_TAC(srw_ss())[DISJOINT_DEF, INTER_DEF, varset_action_def, EXTENSION,FILTER, MEM, SUBSET_DEF]
THEN SRW_TAC[][]
THEN METIS_TAC[]);

val child_parent_lemma_xxx = store_thm("child_parent_lemma_xxx",
``!PROB a vs.(varset_action(a, vs)) 
             ==> (DISJOINT  (FDOM (SND a)) ((prob_dom PROB) DIFF vs))``,
SRW_TAC[][varset_action_def, DISJOINT_DEF, INTER_DEF, EXTENSION]
THEN FULL_SIMP_TAC(srw_ss())[SUBSET_DEF]
THEN METIS_TAC[]);

val child_parent_lemma_xxxx = store_thm("child_parent_lemma_xxxx",
``!PROB a vs. (varset_action(a, (prob_dom PROB) DIFF vs)) 
              ==> (DISJOINT  (FDOM (SND a)) vs)``,
SRW_TAC[][varset_action_def, DISJOINT_DEF, INTER_DEF, EXTENSION]
THEN FULL_SIMP_TAC(srw_ss())[SUBSET_DEF]
THEN METIS_TAC[]);

val child_parent_lemma_xxxxx = store_thm("child_parent_lemma_xxxxx",
``!PROB a vs. (varset_action(a, vs)) /\ (FDOM (SND a) <> EMPTY)
==> (~DISJOINT  (FDOM (SND a)) vs)``,
SRW_TAC[][varset_action_def, DISJOINT_DEF, INTER_DEF, EXTENSION]
THEN FULL_SIMP_TAC(srw_ss())[SUBSET_DEF]
THEN METIS_TAC[]);

val child_parent_lemma_xxxxxx = store_thm("child_parent_lemma_xxxxxx",
``!s a vs.(varset_action(a, vs)) 
          ==> (DISJOINT  (FDOM (SND a)) (s DIFF vs))``,
SRW_TAC[][varset_action_def, DISJOINT_DEF, INTER_DEF, EXTENSION]
THEN FULL_SIMP_TAC(srw_ss())[SUBSET_DEF]
THEN METIS_TAC[]);

val child_parent_lemma_2_1_2_2 = store_thm("child_parent_lemma_2_1_2_2",
``!PROB vs as s. (EVERY (\a. varset_action(a, (prob_dom PROB) DIFF vs)) as)
                  ==>
                   (DRESTRICT s vs = DRESTRICT (exec_plan(s, as)) vs)``,
SRW_TAC[][]
THEN `(∀a. MEM a as  ⇒ DISJOINT (FDOM (SND a)) vs)` by 
      (SRW_TAC[][]
      THEN FULL_SIMP_TAC(srw_ss())[EVERY_MEM]
      THEN METIS_TAC[child_parent_lemma_xxxx])
THEN `EVERY (\a. DISJOINT (FDOM (SND a)) vs) as` by FULL_SIMP_TAC(srw_ss())[EVERY_MEM ]
THEN `EVERY (\a. DISJOINT (FDOM (SND a)) vs) (rem_condless_act(s, [], as))` by FULL_SIMP_TAC(srw_ss())[graph_plan_lemma_7_7]
THEN FULL_SIMP_TAC(srw_ss())[EVERY_MEM]
THEN SRW_TAC[][]
THEN `(exec_plan(s, as) = exec_plan(s, rem_condless_act(s, [], as))) /\ sat_precond_as(s, rem_condless_act(s, [], as))` by SRW_TAC[][Once graph_plan_lemma_7_1, graph_plan_lemma_7_2]
THEN MP_TAC(graph_plan_lemma_4 |>INST_TYPE [beta |-> ``:'d``] |> Q.SPECL [`s`, `s`, `rem_condless_act(s, [], as)`, `vs`, `(\a. varset_action(a, (prob_dom PROB) DIFF vs))`])
THEN SRW_TAC[][]
THEN `EVERY (\a. varset_action (a, (prob_dom PROB) DIFF vs)) (rem_condless_act (s,[],as))` by FULL_SIMP_TAC(srw_ss())[EVERY_MEM, graph_plan_lemma_7_7] 
THEN `FILTER (λa. ¬varset_action (a,(prob_dom PROB) DIFF vs)) (rem_condless_act (s,[],as)) = []` by SRW_TAC[][child_parent_lemma_2_1_2_2_2]
THEN `sat_precond_as (s, FILTER (λa. ¬varset_action (a,(prob_dom PROB) DIFF vs)) (rem_condless_act (s,[],as)))` by SRW_TAC[][sat_precond_as_def]
THEN FULL_SIMP_TAC(srw_ss())[exec_plan_def, EVERY_MEM]);

val bound_child_parent_lemma_1_1_1_4_2_1 = store_thm("bound_child_parent_lemma_1_1_1_4_2_1",
``!as. no_effectless_act(as) ==> (FILTER (\a. ~(FDOM (SND a) = EMPTY)) as = as)``,
METIS_TAC[graph_plan_lemma_16_7,((GSYM FILTER_EQ_ID ) |> Q.ISPEC `(λa. FDOM (SND a) ≠ ∅)` |> Q.SPEC `as`)]);

val parent_children_lemma_1_1_1_7_1 = store_thm("parent_children_lemma_1_1_1_7_1",
``!a vs vs' vs''. varset_action (a,vs'' DIFF (vs' ∪ vs))
                  ==> varset_action (a,vs'' DIFF vs)``,
SRW_TAC[][varset_action_def, DIFF_DEF, UNION_DEF, SUBSET_DEF])

val parent_children_lemma_1_1_1_12_2 = store_thm("parent_children_lemma_1_1_1_12_2",
``(∀s vs vs' a.  varset_action(a, s DIFF (vs UNION vs')) 
                   ⇒  varset_action (a, s DIFF vs'))``,
SRW_TAC[][varset_action_def,
          DISJOINT_DEF, INTER_DEF, EXTENSION,
          FDOM_DRESTRICT, SUBSET_DEF, UNION_DEF]
THEN METIS_TAC[])

val parent_children_main_lemma_1_1_10 = store_thm("parent_children_main_lemma_1_1_10",
``!PROB as. as IN valid_plans PROB
                 ==> (FILTER (\a. varset_action(a, prob_dom PROB)) as = as)``,
STRIP_TAC
THEN Induct_on `as`
THEN SRW_TAC[][]
THEN1 METIS_TAC[valid_plan_valid_tail, valid_plan_valid_head]
THEN FULL_SIMP_TAC(srw_ss())[varset_action_def]
THEN METIS_TAC[valid_plan_valid_head, FDOM_eff_subset_prob_dom_pair])

val parent_children_main_lemma_1_1_11_1 = store_thm("parent_children_main_lemma_1_1_11_1",
``!a vs vs'. vs SUBSET vs' /\ varset_action(a, vs) ==> varset_action(a, vs')``,
SRW_TAC[][SUBSET_DEF, varset_action_def])

val sat_precond_as_pfx = store_thm("sat_precond_as_pfx",
``!s. sat_precond_as (s,as ++ as') ==> 
     sat_precond_as(s,as)``,
Induct_on `as`
THEN SRW_TAC[][sat_precond_as_def])

val _ = export_theory();
