open HolKernel Parse boolLib bossLib;

open finite_mapTheory
open arithmeticTheory
open pred_setTheory
open rich_listTheory
open listTheory; 
open utilsTheory;
open factoredSystemTheory 
open sublistTheory;
open systemAbstractionTheory;
open topologicalPropsTheory;
open list_utilsTheory;
open fmap_utilsTheory;
open actionSeqProcessTheory
open dependencyTheory
open fmap_utilsTheory
open set_utilsTheory
open quantHeuristicsTheory

val _ = new_theory "parentChildStructure";

val child_parent_rel_def 
   = Define `(child_parent_rel(PROB, vs)
     = ~(dep_var_set(PROB, vs,  (prob_dom PROB) DIFF vs)))`;

val child_parent_lemma_1_1_1 = store_thm("child_parent_lemma_1_1_1",
``! a vs. varset_action (a,vs) ==>  (DRESTRICT (SND a) vs = SND a )``,
SRW_TAC[][varset_action_def]
THEN SRW_TAC[][graph_plan_lemma_2_3_8_5]);

val child_parent_lemma_1_1_2 = store_thm("child_parent_lemma_1_1_2",
``! PROB a vs.(a IN PROB /\ child_parent_rel(PROB, vs) /\ (FDOM (SND a) <> EMPTY))
              ==>
                (varset_action(a, vs) <=>
                       ~varset_action(a, (prob_dom PROB) DIFF vs))``,
SRW_TAC[][]
THEN MP_TAC(FDOM_eff_subset_prob_dom_pair |> Q.SPEC `a`)
THEN SRW_TAC[][]
THEN FULL_SIMP_TAC(srw_ss())[varset_action_def, dep_var_set_def, 
                        dep_def, SUBSET_DEF, DISJOINT_DEF,
                        INTER_DEF, EXTENSION, child_parent_rel_def]
THEN PROVE_TAC[]);

val child_parent_lemma_1_1_4 = store_thm("child_parent_lemma_1_4_1",
``!PROB a vs. (child_parent_rel(PROB, vs)) /\ (a IN PROB) /\ ~(DISJOINT (FDOM (SND a)) ((prob_dom PROB) DIFF vs))
==> (DISJOINT  (FDOM (SND a)) vs)``,
SRW_TAC[][]
THEN FULL_SIMP_TAC(srw_ss())[dep_var_set_def, dep_def, DISJOINT_DEF, INTER_DEF, EXTENSION, child_parent_rel_def, SUBSET_DEF]
THEN PROVE_TAC[]);

val as_proj_child_def = Define `as_proj_child(as, vs) = 
    (FILTER (λa. varset_action (a,vs) /\ (FDOM (SND a)) <> EMPTY )(MAP (λa. (DRESTRICT (FST a) vs,SND a)) as))`; 

val child_parent_lemma_1_1 = store_thm("child_parent_lemma_1_1",
``∀PROB as vs.
     (as IN (valid_plans PROB) /\ child_parent_rel(PROB, vs))
     ==> (as_proj_child(as,vs) = as_proj(as,vs))``,
Induct_on `as`
THEN SRW_TAC[][as_proj_child_def, as_proj_def] 
THEN1(FULL_SIMP_TAC(srw_ss())[child_parent_lemma_1_1_1, varset_action_def, action_proj_def])
THEN1(METIS_TAC[as_proj_child_def, as_proj_def, valid_plan_valid_tail])
THEN1(FULL_SIMP_TAC(srw_ss())[child_parent_lemma_1_1_1, varset_action_def])
THEN1(FULL_SIMP_TAC(srw_ss())[]
      THEN1(`~varset_action(h,vs)` by FULL_SIMP_TAC(srw_ss())[varset_action_def]
            THEN METIS_TAC[child_parent_lemma_1_1_4, child_parent_lemma_1_1_5, DISJOINT_DEF, FDOM_DRESTRICT, child_parent_lemma_1_1_3, valid_plan_valid_tail, valid_plan_valid_head])
      THEN1(METIS_TAC[child_parent_lemma_1_1_4, child_parent_lemma_1_1_5, DISJOINT_DEF, FDOM_DRESTRICT, child_parent_lemma_1_1_3, valid_plan_valid_tail, valid_plan_valid_head]))
THEN1 METIS_TAC[as_proj_child_def, as_proj_def, valid_plan_valid_tail, valid_plan_valid_head]);

val graph_plan_lemma_3 = store_thm("graph_plan_lemma_3",
``!vs as. 
no_effectless_act(as) ==>
( LENGTH(FILTER (\a. varset_action(a, vs) ) as) 
	  = LENGTH(as_proj_child(as, vs)))``,
Induct_on`as`
THEN SRW_TAC[][FILTER, as_proj_child_def, no_effectless_act_def]
THEN FULL_SIMP_TAC(srw_ss())[LENGTH_MAP, varset_action_def]);

val graph_plan_lemma_18_1 = store_thm("graph_plan_lemma_18_1",
``!as vs. (as_proj_child(as, vs)) =  (FILTER (\a. varset_action(a, vs) /\ FDOM (SND a) ≠ ∅) (MAP (\ a. (DRESTRICT (FST a) vs, (SND a) )) as))``,
Induct_on `as` 
THEN SRW_TAC[][as_proj_child_def, action_proj_def]);

val graph_plan_lemma_18 = store_thm("graph_plan_lemma_18",
``!as vs. (as_proj_child(rem_effectless_act(as), vs)) =  FILTER (\a. varset_action(a, vs)) (MAP (\ a. (DRESTRICT (FST a) vs, (SND a) )) (rem_effectless_act(as)))``,
Induct_on `as` 
THEN FULL_SIMP_TAC(srw_ss())[as_proj_child_def, action_proj_def, rem_effectless_act_def]
THEN SRW_TAC[][]);

val as_proj_parent_def = Define `as_proj_parent(as, vs) = 
    (FILTER (λa. ~varset_action (a,vs) /\ (FDOM (SND a)) <> EMPTY ) as)`; 

val child_parent_lemma_1_2_1_1 = store_thm("child_parent_lemma_1_2_1_1",
``!PROB vs a. (child_parent_rel(PROB, vs) /\ a IN PROB /\  ~ (DISJOINT (FDOM(SND a)) ((prob_dom PROB) DIFF vs))) 
					  ==>  DISJOINT (FDOM(FST a)) vs``,
SRW_TAC[][]
THEN MP_TAC(FDOM_pre_subset_prob_dom_pair |> Q.SPEC `a`)
THEN SRW_TAC[][]
THEN FULL_SIMP_TAC(srw_ss())[child_parent_rel_def, dep_var_set_def, dep_def, DISJOINT_DEF, DIFF_DEF, EXTENSION, SUBSET_DEF]
THEN PROVE_TAC[]);

val child_parent_lemma_1_2_1 = store_thm("child_parent_lemma_1_2_1",
``!PROB vs a. (child_parent_rel(PROB, vs) /\ a IN PROB /\ varset_action(a, (prob_dom PROB) DIFF vs) 
							         /\ (FDOM (SND a) <> EMPTY))
                    ==> ((FDOM (FST a)) SUBSET (prob_dom PROB) DIFF vs)``,
SRW_TAC[][]
THEN MP_TAC(FDOM_pre_subset_prob_dom_pair |> Q.SPEC `a`)
THEN SRW_TAC[][]
THEN `~DISJOINT (FDOM (SND a)) ((prob_dom PROB) DIFF vs)` by 
     (FULL_SIMP_TAC(srw_ss())[child_parent_rel_def, dep_var_set_def, dep_def, SUBSET_DEF, DISJOINT_DEF, INTER_DEF, varset_action_def, EXTENSION]
     THEN PROVE_TAC[])
THEN `DISJOINT (FDOM (FST a)) vs` by METIS_TAC[child_parent_lemma_1_2_1_1]
THEN METIS_TAC[SUBSET_DIFF]);

val child_parent_lemma_1_2 = store_thm("child_parent_lemma_1_2",
``∀PROB as vs.
     (as IN (valid_plans PROB) /\ child_parent_rel(PROB, vs))
      ==>
         (as_proj_parent(as,vs) = as_proj(as,(prob_dom PROB) DIFF vs))``,
Induct_on `as`
THEN SRW_TAC[][as_proj_parent_def, as_proj_def]
THENL
[
  `varset_action (h,(prob_dom PROB) DIFF vs)` by METIS_TAC[child_parent_lemma_1_1_2, valid_plan_valid_head]
					      THEN ASSUME_TAC(child_parent_lemma_1_2_1 |> Q.SPECL[`PROB`, `vs`, `h`])
					      THEN SRW_TAC[][action_proj_def]
					      THEN FULL_SIMP_TAC(srw_ss())[varset_action_def]
        THEN `h IN PROB` by METIS_TAC[valid_plan_valid_head]
        THEN `FDOM (FST h) ⊆ prob_dom PROB DIFF vs` by METIS_TAC[valid_plan_valid_head]
	THEN SRW_TAC[][graph_plan_lemma_2_3_8_5]
	,
	  METIS_TAC[as_proj_parent_def, as_proj_def, valid_plan_valid_head, valid_plan_valid_tail]
		   ,
		     `varset_action (h,(prob_dom PROB) DIFF vs)` by METIS_TAC[child_parent_lemma_1_1_2, valid_plan_valid_head]
								 THEN FULL_SIMP_TAC(srw_ss())[varset_action_def, FDOM_DRESTRICT, INTER_DEF, SUBSET_DEF, EXTENSION]
								 THEN METIS_TAC[valid_plan_valid_head, valid_plan_valid_tail]
								 ,
								   FULL_SIMP_TAC(srw_ss())[varset_action_def, FDOM_DRESTRICT, INTER_DEF, SUBSET_DEF, EXTENSION]
										THEN METIS_TAC[]
								 ,
        FULL_SIMP_TAC(srw_ss())[as_proj_parent_def]
        THEN METIS_TAC[valid_plan_valid_head, valid_plan_valid_tail]
]);

val child_parent_lemma_1 = store_thm("child_parent_lemma_1",
``∀PROB as vs.
     (as IN (valid_plans PROB) /\ child_parent_rel(PROB, vs))
     ==>
     ((as_proj_child(as,vs) = as_proj(as,vs))
      /\ (as_proj_parent(as,vs) = as_proj(as,(prob_dom PROB) DIFF vs)))``,
METIS_TAC[child_parent_lemma_1_1, child_parent_lemma_1_2]);

val child_parent_lemma_xx = store_thm("child_parent_lemma_xx",
``!PROB a vs. (child_parent_rel(PROB, vs) /\ (a IN PROB) /\ ~(varset_action(a, vs))) 
              ==> (DISJOINT  (FDOM (SND a)) vs)``,
SRW_TAC[][]
THEN MP_TAC(FDOM_eff_subset_prob_dom_pair |> Q.SPEC `a`)
THEN FULL_SIMP_TAC(srw_ss())[SUBSET_DEF, varset_action_def,dep_var_set_def, dep_def, DISJOINT_DEF, INTER_DEF, EXTENSION, child_parent_rel_def]
THEN PROVE_TAC[]);

val child_parent_lemma_2_1_1_1_1_1 = store_thm("child_parent_lemma_2_1_1_1_1_1",
``! PROB s as vs. ( (as IN (valid_plans PROB))  /\ child_parent_rel(PROB, vs)) 
                  ==>
                     ((DRESTRICT s vs) =
                       (DRESTRICT (exec_plan (s,as_proj_parent(as, vs))) vs))``,
Induct_on `as`
THEN SRW_TAC[][as_proj_parent_def, exec_plan_def]
THEN REWRITE_TAC[GSYM as_proj_parent_def]
THEN METIS_TAC[child_parent_lemma_1_1_5, graph_plan_lemma_1_2,
               child_parent_lemma_1_1_4, valid_plan_valid_tail,
               valid_plan_valid_head]);

val child_parent_lemma_2_1_1_1_1 = store_thm("child_parent_lemma_2_1_1_1_1",
``!PROB s s' as vs. ((as IN (valid_plans PROB)) /\ child_parent_rel(PROB, vs)
						        /\ ((DRESTRICT s vs) = (DRESTRICT (exec_plan(s,as)) vs))) 
					        ==>
            ((DRESTRICT s vs) = (DRESTRICT (exec_plan (s,as_proj_parent(as, vs))) vs))``,
METIS_TAC[child_parent_lemma_2_1_1_1_1_1]);

val child_parent_lemma_2_1_1_1_3 = store_thm("child_parent_lemma_2_1_1_1_3",
``!as vs. (?a. MEM a as /\ ((\a. varset_action(a,vs)) a))
               ==> LENGTH
                      (FILTER (λa. varset_action (a,vs))
                              (as_proj_parent(as, vs)))
                   < LENGTH (FILTER (λa. varset_action (a,vs)) as)``,
SRW_TAC[][as_proj_parent_def]
THEN METIS_TAC[Q.SPEC ` as` ( Q.SPEC  `(\a. FDOM (SND a) ≠ ∅) ` (Q.ISPEC `(\a. varset_action(a, vs))` child_parent_lemma_2_1_1_1_3_1))]);

val child_parent_lemma_2_1_1_1 = store_thm("child_parent_lemma_2_1_1_1",
``!PROB s as vs. ((as IN (valid_plans PROB)) /\ child_parent_rel(PROB, vs)
                  /\ ((DRESTRICT s vs) = (DRESTRICT (exec_plan(s,as)) vs))
                  /\ (?a. MEM a as /\ ((\a. varset_action(a,vs)) a))) 
					      ==>
            (((DRESTRICT (exec_plan(s,as)) vs) = (DRESTRICT (exec_plan (s,as_proj_parent(as, vs))) vs)) /\ (LENGTH( FILTER (\a. varset_action(a,vs)) (as_proj_parent(as, vs))) < LENGTH(FILTER (\a. varset_action(a,vs)) as)) )``,
METIS_TAC[child_parent_lemma_2_1_1_1_1, child_parent_lemma_2_1_1_1_3]);

val child_parent_lemma_2_1_1_2 = store_thm("child_parent_lemma_2_1_1_2",
``!PROB s as vs. ((as IN (valid_plans PROB)) /\ child_parent_rel(PROB, vs) /\ sat_precond_as (s,as))
                 ==>
                 (sat_precond_as (s,as_proj_parent(as, vs)))``,
METIS_TAC[graph_plan_lemma_12,  child_parent_lemma_1]);

val child_parent_lemma_2_1_2_1 = store_thm("child_parent_lemma_2_1_2_1",
``!PROB vs as. (child_parent_rel(PROB, vs) /\ as IN (valid_plans PROB))
               ==>
                  (as_proj_parent (as,vs) = FILTER (λa. varset_action (a, (prob_dom PROB) DIFF vs) ∧ FDOM (SND a) ≠ ∅) as)``,
Induct_on `as`
THEN SRW_TAC[][as_proj_parent_def]
THEN SRW_TAC[][GSYM as_proj_parent_def]
THEN FULL_SIMP_TAC(srw_ss())[GSYM as_proj_parent_def, child_parent_lemma_1_1_2]
THEN METIS_TAC[child_parent_lemma_1_1_2, valid_plan_valid_head, valid_plan_valid_tail]);

val child_parent_lemma_2_1_2_4 = store_thm("child_parent_lemma_2_1_2_4",
``! PROB as vs . (as IN (valid_plans PROB) /\ child_parent_rel(PROB, vs))
                 ==>
                    (FILTER (λa. varset_action (a,(prob_dom PROB) DIFF vs) /\ FDOM (SND a) ≠ ∅ ) as  = 
                        FILTER (λa. ¬varset_action (a,vs) ∧ FDOM (SND a) ≠ ∅) as)``,
Induct_on `as`
THEN1 SRW_TAC[][]
THEN REPEAT GEN_TAC
THEN STRIP_TAC 
THEN SRW_TAC[][]
THEN FULL_SIMP_TAC(srw_ss())[]
THEN METIS_TAC[child_parent_lemma_1_1_2, valid_plan_valid_tail, valid_plan_valid_head]);

fun cheat g = ACCEPT_TAC (mk_thm g) g

val replace_projected_def = 
   Define `(replace_projected(as'', a'::as', a::as, vs) = 
	    if (varset_action(a, vs)) then 
		   if (a' = (action_proj(a, vs))) then
		             replace_projected(as'' ++ [a], as', as, vs)
					         else 
						           replace_projected(as'', a'::as', as, vs)
									    else
										   replace_projected(as'' ++ [a], a'::as', as, vs))
           /\ ((replace_projected(as'', [], as, vs)) =  as'' ++ (FILTER (\a. ~varset_action(a, vs)) as))
	         /\ ((replace_projected(as'', as, [], vs)) =  as'')`;

val stitch_def = Define`
  (stitch (a'::as') vs (a::as) = if varset_action(a, vs) then 
                                   if a' = action_proj(a,vs) then 
                                     a::stitch as' vs as
                                   else stitch (a'::as') vs as
                                 else a::stitch (a'::as') vs as) /\
  (stitch [] vs as = FILTER (\a. ~varset_action (a, vs)) as) /\
  (stitch (a'::as') vs  [] = [])`;

val replace_projected_stitch = store_thm(
  "replace_projected_stitch",
  ``!as as' acc. replace_projected (acc, as', as, vs) = acc ++ stitch as' vs as``,
  Induct_on `as` THEN Cases_on `as'` THEN 
  SRW_TAC [][stitch_def, replace_projected_def]);

val graph_plan_lemma_22_1 = store_thm("graph_plan_lemma_22_1",
``!as as' as'' vs s. set as SUBSET s  /\ set as'' SUBSET s
					      ==> 
					       set (replace_projected(as'', as', as, vs)) SUBSET s``,
Induct_on `as`
THEN Cases_on `as'`
THEN SRW_TAC[][replace_projected_def]
THEN METIS_TAC[graph_plan_lemma_6_2]);

val graph_plan_lemma_22_1' = store_thm("graph_plan_lemma_22_1'",
``!as as' as'' vs s. as IN (valid_plans s)  /\ as'' IN (valid_plans s)
					            ==> 
						     (replace_projected(as'', as', as, vs)) IN (valid_plans s)``,
Induct_on `as`
THEN Cases_on `as'`
THEN FULL_SIMP_TAC(srw_ss())[replace_projected_def, valid_plans_def]
THEN SRW_TAC[][]
THEN METIS_TAC[graph_plan_lemma_6_2]);

val graph_plan_lemma_22_2_1 = store_thm("graph_plan_lemma_22_2_1",
``! as as' as'' as''' as'''' vs. sublist (replace_projected (as'',as',as,vs)) as'''
				         ==> sublist (replace_projected (as'''' ++ as'',as',as,vs)) (as'''' ++as''')``,
Induct_on `as`
THENL
[
  Cases_on `as'`
	   THEN SRW_TAC[][replace_projected_def, sublist_filter, sublist_cons, sublist_trans, sublist_refl]
	   THEN METIS_TAC[sublist_refl, sublist_append, GSYM APPEND_ASSOC]
	   ,
	     SRW_TAC[][]
		    THEN Cases_on `as'`
		    THEN SRW_TAC[][replace_projected_def, sublist_filter, sublist_cons, sublist_trans, sublist_def]
		    THEN FULL_SIMP_TAC(srw_ss())[replace_projected_def, sublist_filter, sublist_cons, sublist_trans, sublist_def]
		    THEN METIS_TAC[sublist_refl, sublist_append, GSYM APPEND_ASSOC]
]);

val graph_plan_lemma_22_2 = store_thm("graph_plan_lemma_22_2",
``!as as' vs s. sublist (replace_projected([], as', as, vs)) as``,
Induct_on `as`
THEN Cases_on `as'`
THEN SRW_TAC[][replace_projected_def, sublist_filter, sublist_cons, sublist_trans, sublist_def]
THEN MP_TAC(graph_plan_lemma_22_2_1
		|> Q.SPEC `as`
		|> Q.SPEC `t`
		|> Q.SPEC `[]`
		|> Q.SPEC `as`
		|> Q.SPEC `[h']`
		|> Q.SPEC `vs`)
THEN MP_TAC(graph_plan_lemma_22_2_1
		|> Q.SPEC `as`
		|> Q.SPEC `h::t`
		|> Q.SPEC `[]`
		|> Q.SPEC `as`
		|> Q.SPEC `[h']`
		|> Q.SPEC `vs`)
THEN SRW_TAC[][sublist_def]);

val graph_plan_lemma_22_3_1 = store_thm("graph_plan_lemma_22_3_1",
``!as''' as'' as' as vs. replace_projected (as''' ++ as'', as', as, vs) = as''' ++ replace_projected (as'', as', as, vs)``,
Induct_on `as`
THENL
[
  Cases_on `as'`
	   THEN SRW_TAC[][replace_projected_def, sublist_filter, sublist_cons, sublist_trans, sublist_refl]
	   THEN METIS_TAC[sublist_refl, sublist_append, GSYM APPEND_ASSOC]
	   ,
	     SRW_TAC[][]
		    THEN Cases_on `as'`
		    THEN SRW_TAC[][replace_projected_def, sublist_filter, sublist_cons, sublist_trans, sublist_def]
]);

val graph_plan_lemma_22_3 = store_thm("graph_plan_lemma_22_3",
``!PROB vs as' as. no_effectless_act(as) /\  as IN (valid_plans PROB)
					        ==> no_effectless_act(replace_projected([], as', as, vs))``,
NTAC 2 STRIP_TAC
THEN Induct_on `as`
THEN Cases_on `as'`
THEN SRW_TAC[SatisfySimps.SATISFY_ss][no_effectless_act_def, replace_projected_def, graph_plan_lemma_16_10]
THEN FULL_SIMP_TAC(bool_ss)[]
THEN MP_TAC(graph_plan_lemma_22_3_1
		    |> Q.SPEC `[h']`
		        |> Q.SPEC `[]`
			    |> Q.SPEC `t`
			        |> Q.SPEC `as`
				    |> Q.SPEC `vs`) 
THEN SRW_TAC[][]
THEN SRW_TAC[][no_effectless_act_def]
THEN MP_TAC(child_parent_lemma_xxxxx
		  |> Q.SPEC `PROB`
		    |> Q.SPEC `h'`
		      |> Q.SPEC `vs`)
THEN SRW_TAC[][]
THEN MP_TAC(REWRITE_RULE[FDOM_DRESTRICT, GSYM DISJOINT_DEF] (graph_plan_lemma_2_3_9
								   |> INST_TYPE [alpha |-> ``:'a``] 
								     |> INST_TYPE [beta |-> ``:bool``]
								       |> INST_TYPE [gamma |-> ``:bool``]
								         |> Q.SPEC `as`
									   |> Q.SPEC `h'`
									     |> Q.SPEC `vs`))
THEN SRW_TAC[SatisfySimps.SATISFY_ss][]
THEN FULL_SIMP_TAC(bool_ss)[sublist_cons_2]
THEN FULL_SIMP_TAC(srw_ss())[sublist_def]
THEN SRW_TAC[][]
THEN MP_TAC(graph_plan_lemma_22_3_1
		    |> Q.SPEC `[h']`
		        |> Q.SPEC `[]`
			    |> Q.SPEC `h::t`
			        |> Q.SPEC `as`
				    |> Q.SPEC `vs`) 
THEN SRW_TAC[][]
THEN SRW_TAC[][no_effectless_act_def]
THEN METIS_TAC[valid_plan_valid_tail]);

val graph_plan_lemma_22_4 = store_thm("graph_plan_lemma_22_4",
``!PROB vs as as'. (as) IN (valid_plans PROB) /\ child_parent_rel(PROB, vs) /\ no_effectless_act(as)
									             ==> ((replace_projected([], as', as, vs))) IN (valid_plans PROB)``,
NTAC 3 STRIP_TAC
THEN Induct_on `as`
THEN SRW_TAC[][]
THEN Cases_on `as'`
THEN SRW_TAC[SatisfySimps.SATISFY_ss][no_effectless_act_def, replace_projected_def, graph_plan_lemma_16_10]
THEN SRW_TAC[SatisfySimps.SATISFY_ss][graph_plan_lemma_6_2]
THEN MP_TAC(graph_plan_lemma_22_3_1
		    |> Q.SPEC `[h]`
		        |> Q.SPEC `[]`
			    |> Q.SPEC `t`
			        |> Q.SPEC `as`
				    |> Q.SPEC `vs`) 
THEN SRW_TAC[][]
THEN FULL_SIMP_TAC(bool_ss)[no_effectless_act_def]
THEN MP_TAC(child_parent_lemma_xxxxx
		  |> Q.SPEC `PROB`
		    |> Q.SPEC `h`
		      |> Q.SPEC `vs`)
THEN SRW_TAC[][]
THEN MP_TAC(REWRITE_RULE[FDOM_DRESTRICT, GSYM DISJOINT_DEF] (graph_plan_lemma_2_3_9
								   |> INST_TYPE [alpha |-> ``:'a``] 
								     |> INST_TYPE [beta |-> ``:bool``]
								       |> INST_TYPE [gamma |-> ``:bool``]
								         |> Q.SPEC `as`
									   |> Q.SPEC `h`
									     |> Q.SPEC `vs`))
THEN SRW_TAC[SatisfySimps.SATISFY_ss][]
THEN FULL_SIMP_TAC(bool_ss)[sublist_cons_2]
THEN FULL_SIMP_TAC(srw_ss())[sublist_def]
THEN MP_TAC(graph_plan_lemma_22_3_1
		    |> Q.SPEC `[h]`
		        |> Q.SPEC `[]`
			    |> Q.SPEC `h'::t`
			        |> Q.SPEC `as`
				    |> Q.SPEC `vs`) 
THEN SRW_TAC[][]
THEN FULL_SIMP_TAC(bool_ss)[no_effectless_act_def]
THEN MP_TAC(child_parent_lemma_xx
		  |> Q.SPEC `PROB`
		    |> Q.SPEC `h`
		      |> Q.SPEC `vs`)
THEN SRW_TAC[][]
THEN MP_TAC (REWRITE_RULE[FDOM_DRESTRICT, GSYM DISJOINT_DEF] (graph_plan_lemma_2_3_10
								    |> INST_TYPE [alpha |-> ``:'a``] 
								      |> INST_TYPE [beta |-> ``:bool``]
								        |> INST_TYPE [gamma |-> ``:bool``]
									  |> Q.SPEC `as`
									    |> Q.SPEC `h`
									      |> Q.SPEC `vs`))
THEN SRW_TAC[SatisfySimps.SATISFY_ss][]
THEN FULL_SIMP_TAC(bool_ss)[sublist_def]
THEN METIS_TAC[valid_head_and_tail_valid_plan, valid_plan_valid_tail, valid_plan_valid_head, graph_plan_lemma_22_4_1]);

val bound_child_parent_lemma_1_1_1_2_1_1 = store_thm("bound_child_parent_lemma_1_1_1_2_1_1",
``!PROB as vs. no_effectless_act(as) /\ child_parent_rel(PROB, vs) /\ as IN (valid_plans PROB)
								             ==> ((as_proj (FILTER (λa. ¬varset_action (a,vs)) as,vs)) = []) ``,
Induct_on `as`
THEN SRW_TAC[][as_proj_def, no_effectless_act_def]
THENL[
    REWRITE_TAC[FDOM_DRESTRICT]
	       THEN REWRITE_TAC[GSYM DISJOINT_DEF]
	       THEN MP_TAC(child_parent_lemma_xx
			         |> Q.SPEC `PROB`
				   |> Q.SPEC `h`
				     |> Q.SPEC `vs`)
        THEN SRW_TAC[][]
	THEN METIS_TAC[valid_plan_valid_head, valid_plan_valid_tail]
	,
	  METIS_TAC[valid_plan_valid_head, valid_plan_valid_tail]
]);

val bound_child_parent_lemma_1_1_1_2_1 = store_thm("bound_child_parent_lemma_1_1_1_2_1",
``!PROB as as' vs. no_effectless_act(as) /\ child_parent_rel(PROB, vs) /\ as IN (valid_plans PROB) /\ sublist as' (as_proj(as, vs))
												          ==> (as' = (as_proj(replace_projected([], as', as, vs), vs)))``,
Induct_on `as`
THEN SRW_TAC[][as_proj_def, sublist_def, replace_projected_def]
THEN Cases_on `as'`
THEN SRW_TAC[][as_proj_def, sublist_def, replace_projected_def]
THEN FULL_SIMP_TAC(bool_ss)[no_effectless_act_def]
THENL[
          FULL_SIMP_TAC(srw_ss())[FDOM_DRESTRICT]
		         THEN FULL_SIMP_TAC(srw_ss())[GSYM DISJOINT_DEF]
			   THEN METIS_TAC[child_parent_lemma_xx, valid_plan_valid_head, valid_plan_valid_tail]
	,    
          METIS_TAC[bound_child_parent_lemma_1_1_1_2_1_1, valid_plan_valid_head, valid_plan_valid_tail]
		     ,
		         REWRITE_TAC[REWRITE_RULE[APPEND_NIL](graph_plan_lemma_22_3_1
								  |> Q.SPEC `[h]`
								  |> Q.SPEC `[]`
								  |> Q.SPEC `t`
								  |> Q.SPEC `as`
								  |> Q.SPEC `vs`)]
				      THEN SRW_TAC[][]
				        THEN MP_TAC(graph_plan_lemma_2_3_9 
							  |> Q.ISPEC `(replace_projected ([],t,as,vs))`
							    |> Q.SPEC `h`
							      |> Q.SPEC `vs`)
					  THEN SRW_TAC[][]
					    THEN FULL_SIMP_TAC(bool_ss)[sublist_cons_2, no_effectless_act_def]
					      THEN METIS_TAC[valid_plan_valid_head, valid_plan_valid_tail]
					        ,
						    FULL_SIMP_TAC(bool_ss)[sublist_def, no_effectless_act_def]
								   THEN METIS_TAC[valid_plan_valid_head, valid_plan_valid_tail]
								     ,
          FULL_SIMP_TAC(srw_ss())[FDOM_DRESTRICT]
		         THEN FULL_SIMP_TAC(srw_ss())[GSYM DISJOINT_DEF]
			   THEN METIS_TAC[child_parent_lemma_xx, valid_plan_valid_head, valid_plan_valid_tail] 
			     ,
          METIS_TAC[bound_child_parent_lemma_1_1_1_2_1_1, valid_plan_valid_head, valid_plan_valid_tail] 
		     ,
          METIS_TAC[bound_child_parent_lemma_1_1_1_2_1_1, valid_plan_valid_head, valid_plan_valid_tail]
		     ,
		         FULL_SIMP_TAC(bool_ss)[no_effectless_act_def]
          THEN FULL_SIMP_TAC(srw_ss())[FDOM_DRESTRICT]
	    THEN FULL_SIMP_TAC(srw_ss())[GSYM DISJOINT_DEF]
	      THEN METIS_TAC[child_parent_lemma_xxxxx, valid_plan_valid_head, valid_plan_valid_tail] 
	        ,
		    FULL_SIMP_TAC(bool_ss)[no_effectless_act_def]
          THEN FULL_SIMP_TAC(srw_ss())[FDOM_DRESTRICT]
	    THEN FULL_SIMP_TAC(srw_ss())[GSYM DISJOINT_DEF]
	      THEN METIS_TAC[child_parent_lemma_xxxxx, valid_plan_valid_head, valid_plan_valid_tail]
	        ,
		    REWRITE_TAC[REWRITE_RULE[APPEND_NIL](graph_plan_lemma_22_3_1
							     |> Q.SPEC `[h]`
							     |> Q.SPEC `[]`
							     |> Q.SPEC `h'::t`
							     |> Q.SPEC `as`
							     |> Q.SPEC `vs`)]
			         THEN SRW_TAC[][]
				   THEN MP_TAC(graph_plan_lemma_2_3_10
						     |> Q.ISPEC `(replace_projected ([],h'::t,as,vs))`
						       |> Q.SPEC `h`
						         |> Q.SPEC `vs`)
				     THEN SRW_TAC[][]
				       THEN FULL_SIMP_TAC(bool_ss)[sublist_cons_2, no_effectless_act_def]
				         THEN METIS_TAC[valid_plan_valid_head, valid_plan_valid_tail]
]);

val bound_child_parent_lemma_1_1_1_2 = store_thm("bound_child_parent_lemma_1_1_1_2",
``!PROB s as vs as'. child_parent_rel(PROB, vs) /\ (s IN (valid_states PROB))
						    /\ as IN (valid_plans PROB) /\ sublist as' (as_proj(as, vs)) /\ no_effectless_act(as)
														     ==>
														         (LENGTH as' = LENGTH (FILTER (\a. varset_action(a, vs)) (replace_projected([], as', as, vs))))``,
SRW_TAC[][]
THEN MP_TAC(bound_child_parent_lemma_1_1_1_2_1
		   |> Q.SPEC `PROB`
		      |> Q.SPEC `as`
		         |> Q.SPEC `as'`
			    |> Q.SPEC `vs`)
THEN SRW_TAC[][]
THEN MP_TAC(graph_plan_lemma_22_4
		|> Q.SPEC `PROB`
		|> Q.SPEC `vs`
		|> Q.SPEC `as`
		|> Q.SPEC `as'`)
THEN SRW_TAC[][]
THEN MP_TAC(child_parent_lemma_1_1
		  |> Q.SPEC `PROB`
		    |> Q.SPEC `(replace_projected ([],as',as,vs))`
		      |> Q.SPEC `vs`)
THEN SRW_TAC[][]
THEN MP_TAC(graph_plan_lemma_22_3
		|> Q.SPEC `PROB`
		|> Q.SPEC `vs`
		|> Q.SPEC `as'`
		|> Q.SPEC `as`)
THEN SRW_TAC[][]
THEN MP_TAC(graph_plan_lemma_3
		|> Q.SPEC `vs`
		|> Q.SPEC `(replace_projected ([],as',as,vs))`)
THEN SRW_TAC[][]
THEN METIS_TAC[]);

val bound_child_parent_lemma_1_1_1_3_1_1 = store_thm("bound_child_parent_lemma_1_1_1_3_1_1",
``!PROB vs as s s'. as IN (valid_plans PROB) /\ child_parent_rel(PROB, vs)
						 /\ sat_precond_as(s, as) /\ (s IN (valid_states PROB))
									      /\ (DRESTRICT s ((prob_dom PROB) DIFF vs) = DRESTRICT s' ((prob_dom PROB) DIFF vs))
										  /\ no_effectless_act(as)
										      ==> sat_precond_as (s',FILTER (λa. ¬varset_action (a,vs)) as)``,
SRW_TAC[][]
THEN `EVERY (λa. FDOM (SND a) ≠ ∅) as ` by METIS_TAC[
     (graph_plan_lemma_16_7
	  |> Q.SPEC `as`)]
THEN SRW_TAC[][]
THEN MP_TAC(child_parent_lemma_2_1_2_1
                         |> INST_TYPE[gamma |-> ``:'b``]
			  |> Q.SPEC `PROB`
			   |> Q.SPEC `vs`
			    |> Q.SPEC `as`)
THEN SRW_TAC[][]
THEN MP_TAC(EVERY_ABSORB
		|> INST_TYPE [alpha |-> ``:('a |-> 'b) # ('a |-> 'b)``]
		|> Q.SPEC `(λa. varset_action (a, (prob_dom (PROB:(('a |-> 'b) # ('a |-> 'b)) set)) DIFF vs))`
		|> Q.SPEC `(\a. FDOM (SND a) ≠ ∅)`
		|> Q.SPEC `as`)
THEN SRW_TAC[][]
THEN FULL_SIMP_TAC(srw_ss())[]
THEN MP_TAC ((child_parent_lemma_2_1_1_2
		  |> Q.SPEC `PROB`
		  |> Q.SPEC `s`
		  |> Q.SPEC `as`
		  |> Q.SPEC `vs`))
THEN ASM_SIMP_TAC(srw_ss())[]
THEN SRW_TAC[][]
THEN MP_TAC(GSYM (child_parent_lemma_1_2
                  |> INST_TYPE[gamma |-> ``:'b``] 
		    |> Q.SPEC `PROB`
		      |> Q.SPEC `as`
		        |> Q.SPEC `vs`))
THEN SRW_TAC[][]
THEN MP_TAC(graph_plan_lemma_12
		|> Q.SPEC `as`
		|> Q.SPEC `s`
		|> Q.SPEC `s'`
		|> Q.SPEC `(prob_dom (PROB:(('a |-> 'b) # ('a |-> 'b)) set)) DIFF vs`)
THEN SRW_TAC[][]
THEN MP_TAC(EVERY_ABSORB
		|> INST_TYPE [alpha |-> ``:('a |-> 'b) # ('a |-> 'b)``]
		|> Q.SPEC `(λa. ~varset_action (a,vs))`
		|> Q.SPEC `(\a. FDOM (SND a) ≠ ∅)`
		|> Q.SPEC `as`)
THEN SRW_TAC[][]
THEN FULL_SIMP_TAC(srw_ss())[]
THEN METIS_TAC[child_parent_lemma_2_1_2_4 |> INST_TYPE [gamma |-> ``:'b``]]);

val bound_child_parent_lemma_1_1_1_3_1_1 = store_thm("bound_child_parent_lemma_1_1_1_3_1_1",
``!PROB vs a s s'. child_parent_rel(PROB, vs) /\ a IN PROB /\ ~(FDOM (SND a) = EMPTY) /\ (s IN (valid_states PROB)) /\ (s' IN (valid_states PROB))
              /\ (DRESTRICT s ((prob_dom PROB) DIFF vs) = DRESTRICT s' ((prob_dom PROB) DIFF vs))
              /\ ~varset_action(a, vs) /\ FST a SUBMAP s
              ==>
                  FST a SUBMAP s'``,
SRW_TAC[][]
THEN MP_TAC(child_parent_lemma_1_1_2 |> Q.SPEC `PROB` |> Q.SPEC `a` |> Q.SPEC `vs`)
THEN SRW_TAC[][]
THEN MP_TAC(child_parent_lemma_1_2_1|> Q.SPEC `PROB` |> Q.SPEC `vs` |> Q.SPEC `a`)
THEN SRW_TAC[][]
THEN METIS_TAC[graph_plan_lemma_2_3_8_5, bound_child_parent_lemma_1_1_1_3_1_1_1]);

val bound_child_parent_lemma_1_1_1_3_1 = store_thm("bound_child_parent_lemma_1_1_1_3_1",
``!PROB vs sa sc as as_c. sublist as_c (as_proj(as, vs)) /\ (sa IN (valid_states PROB))
           /\ (sc IN (valid_states PROB)) /\ (child_parent_rel(PROB, vs))
           /\ as IN (valid_plans PROB)  /\ no_effectless_act(as) 
           /\ (DRESTRICT sa ((prob_dom PROB) DIFF vs) = DRESTRICT sc ((prob_dom PROB) DIFF vs))
           /\ sat_precond_as(sa, as)
           /\ sat_precond_as(sc, as_c)
           ==>
               (sat_precond_as(sc, (replace_projected([], as_c, as, vs))))``,
NTAC 2 STRIP_TAC
THEN Induct_on `as`
THEN Cases_on `as_c`
THEN(SRW_TAC[][replace_projected_def, sat_precond_as_def, no_effectless_act_def])
THENL
[
   MP_TAC(bound_child_parent_lemma_1_1_1_3_1_1
                      |> INST_TYPE[gamma |-> ``:'b``]
                      |> Q.SPEC `PROB`
                      |> Q.SPEC `vs`
                      |> Q.SPEC `h`
                      |> Q.SPEC `sa`
                      |> Q.SPEC `sc`)
   THEN SRW_TAC[][]
   THEN METIS_TAC[valid_plan_valid_head, valid_plan_valid_tail]
   ,
   ASSUME_TAC(MATCH_MP AND2_THM (MATCH_MP AND1_THM (REWRITE_RULE[CONJ_ASSOC] (replace_projected_def|>INST_TYPE [gamma |-> ``:'b``]))) |> Q.SPEC `vs` |> Q.SPEC `[]` |> Q.SPEC `as`)
   THEN FIRST_X_ASSUM (Q.SPECL_THEN [`state_succ sa h`, `state_succ sc h`, `[]`] MP_TAC)
   THEN SRW_TAC[][]
   THEN FIRST_X_ASSUM MATCH_MP_TAC
   THEN SRW_TAC[][]
   THENL
   [
      METIS_TAC[valid_action_valid_succ, valid_plan_valid_head]
      ,
      METIS_TAC[valid_action_valid_succ, valid_plan_valid_head]
      ,
      METIS_TAC[valid_plan_valid_tail]
      ,
      MATCH_MP_TAC(graph_plan_lemma_4_1)
      THEN SRW_TAC[][]
      THEN MATCH_MP_TAC(bound_child_parent_lemma_1_1_1_3_1_1)
      THEN Q.EXISTS_TAC `PROB`
      THEN Q.EXISTS_TAC `vs`
      THEN Q.EXISTS_TAC `sa`
      THEN SRW_TAC[][]
      THEN METIS_TAC[valid_plan_valid_head]
      ,
      SRW_TAC[][sat_precond_as_def]
   ]
   ,
   ASSUME_TAC(MATCH_MP AND2_THM (MATCH_MP AND1_THM (REWRITE_RULE[CONJ_ASSOC] (replace_projected_def|>INST_TYPE[gamma |-> ``:'b``]))) |> Q.SPECL [`vs`, `[]`, `as`])
   THEN FIRST_X_ASSUM (Q.SPECL_THEN [`state_succ sa h`, `sc`, `[]`] MP_TAC)
   THEN SRW_TAC[][]
   THEN FIRST_X_ASSUM MATCH_MP_TAC
   THEN SRW_TAC[][]
   THENL
   [
      METIS_TAC[valid_action_valid_succ, valid_plan_valid_head]
      ,
      METIS_TAC[valid_plan_valid_tail]
      ,
      METIS_TAC[graph_plan_lemma_1_2, child_parent_lemma_xxx]
      ,
      SRW_TAC[][sat_precond_as_def]
   ]
   ,
   MP_TAC(graph_plan_lemma_22_3_1
                       |> INST_TYPE[gamma |-> ``:'b``]
                       |> Q.SPEC `[h']`
                       |> Q.SPEC `[]`
                       |> Q.SPEC `t`
                       |> Q.SPEC `as`
                       |> Q.SPEC `vs`)
   THEN SRW_TAC[][sat_precond_as_def]
   THENL
   [
      MATCH_MP_TAC(bound_child_parent_lemma_1_1_1_3_1_2)
      THEN Q.EXISTS_TAC `vs`
      THEN SRW_TAC[][]
      THEN1 (FULL_SIMP_TAC(srw_ss())[valid_states_def] THEN METIS_TAC[FDOM_pre_subset_prob_dom_pair, valid_plan_valid_head])
      THEN FULL_SIMP_TAC(srw_ss())[action_proj_def, valid_states_def]
      THEN MATCH_MP_TAC(bound_child_parent_lemma_1_1_1_3_1_1_1)
      THEN Q.EXISTS_TAC `sa`
      THEN SRW_TAC[][]
      THEN METIS_TAC[]
      ,
      FIRST_X_ASSUM MATCH_MP_TAC
      THEN SRW_TAC[][]
      THEN Q.EXISTS_TAC `state_succ sa h'`
      THEN SRW_TAC[][]
      THENL
      [
         FULL_SIMP_TAC(bool_ss)[varset_action_def, FDOM_DRESTRICT, as_proj_def]
         THEN MP_TAC(graph_plan_lemma_2_3_9 |> Q.SPEC `as` |> Q.SPEC `h'` |> Q.SPEC `vs`)
         THEN SRW_TAC[][FDOM_DRESTRICT]
         THEN MP_TAC((bound_child_parent_lemma_1_1_1_3_1_3 |> Q.SPEC `FDOM (SND h')` |> Q.SPEC `vs`) |> INST_TYPE [gamma |-> ``:'a |-> 'b``, alpha |-> ``:'a``, beta |-> ``:'b``])
         THEN SRW_TAC[][]
         THEN METIS_TAC[sublist_cons_2]
         ,
         METIS_TAC[valid_plan_valid_head, valid_action_valid_succ]
         ,
         METIS_TAC[valid_plan_valid_head, valid_action_valid_succ]
         ,
         METIS_TAC[valid_plan_valid_tail]
         ,
         METIS_TAC[child_parent_lemma_xxx, graph_plan_lemma_1_2]
         ,
         `DRESTRICT (FST h') vs ⊑ sc` by FULL_SIMP_TAC(srw_ss())[action_proj_def]
         THEN FULL_SIMP_TAC(srw_ss())[valid_states_def]
         THEN `(FST h') SUBMAP sc`
                by (MATCH_MP_TAC bound_child_parent_lemma_1_1_1_3_1_2
                    THEN Q.EXISTS_TAC `vs`
                    THEN SRW_TAC[][]
                    THEN1(METIS_TAC[valid_plan_valid_head, FDOM_pre_subset_prob_dom_pair])
                    THEN METIS_TAC[bound_child_parent_lemma_1_1_1_3_1_1_1])        
         THEN `state_succ sc (action_proj (h',vs)) = state_succ sc h'`
              by(MATCH_MP_TAC bound_child_parent_lemma_1_1_1_3_1_4
                 THEN SRW_TAC[][]
                 THEN METIS_TAC[bound_child_parent_lemma_1_1_1_3_1_4, valid_plan_valid_head, FDOM_eff_subset_prob_dom_pair])
         THEN METIS_TAC[]
      ]
   ]
   ,
   FIRST_X_ASSUM MATCH_MP_TAC
   THEN Q.EXISTS_TAC `state_succ sa h'`
   THEN SRW_TAC[][]
   THENL
   [
      FULL_SIMP_TAC(bool_ss)[varset_action_def, FDOM_DRESTRICT, as_proj_def]
      THEN MP_TAC(graph_plan_lemma_2_3_9 |> Q.SPEC `as` |> Q.SPEC `h'` |> Q.SPEC `vs`)
      THEN SRW_TAC[][FDOM_DRESTRICT]
      THEN MP_TAC((bound_child_parent_lemma_1_1_1_3_1_3 |> Q.SPEC `FDOM (SND h')` |> Q.SPEC `vs`) |> INST_TYPE [gamma |-> ``:'a |-> 'b``, alpha |-> ``:'a``, beta |-> ``:'b``])
      THEN SRW_TAC[][]
      THEN FULL_SIMP_TAC(srw_ss())[sublist_def]
      THEN METIS_TAC[]
      ,
      METIS_TAC[valid_plan_valid_head, valid_action_valid_succ]
      ,
      METIS_TAC[valid_plan_valid_tail]
      ,
      METIS_TAC[child_parent_lemma_xxx, graph_plan_lemma_1_2]
      ,
      SRW_TAC[][sat_precond_as_def]
   ]
   ,
   MP_TAC(graph_plan_lemma_22_3_1
                       |> INST_TYPE[gamma |-> ``:'b``]
                       |> Q.SPEC `[h']`
                       |> Q.SPEC `[]`
                       |> Q.SPEC `h::t`
                       |> Q.SPEC `as`
                       |> Q.SPEC `vs`)
   THEN SRW_TAC[][sat_precond_as_def]
   THEN1 METIS_TAC[bound_child_parent_lemma_1_1_1_3_1_1, valid_plan_valid_head, FDOM_pre_subset_prob_dom_pair]
   THEN FIRST_X_ASSUM MATCH_MP_TAC
   THEN Q.EXISTS_TAC `state_succ sa h'`
   THEN SRW_TAC[][]
   THENL
   [
      FULL_SIMP_TAC(bool_ss)[as_proj_def, FDOM_DRESTRICT]
      THEN MP_TAC(REWRITE_RULE[DISJOINT_DEF] child_parent_lemma_xx |> INST_TYPE[gamma |-> ``:'b``])
      THEN METIS_TAC[valid_plan_valid_head]
      ,
      METIS_TAC[valid_plan_valid_head, valid_action_valid_succ]
      ,
      METIS_TAC[valid_plan_valid_head, valid_action_valid_succ]
      ,
      METIS_TAC[valid_plan_valid_tail]
      ,
      MATCH_MP_TAC(graph_plan_lemma_4_1)
      THEN SRW_TAC[][]
      THEN MATCH_MP_TAC(bound_child_parent_lemma_1_1_1_3_1_1)
      THEN Q.EXISTS_TAC `PROB`
      THEN Q.EXISTS_TAC `vs`
      THEN Q.EXISTS_TAC `sa`
      THEN SRW_TAC[][]
      THEN METIS_TAC[valid_plan_valid_head]
      ,
      `DRESTRICT (state_succ sc h') vs = DRESTRICT sc vs` by METIS_TAC[child_parent_lemma_xx, graph_plan_lemma_1_2, valid_plan_valid_head]
      THEN METIS_TAC[graph_plan_lemma_13, graph_plan_lemma_12, graph_plan_lemma_23, sat_precond_as_def]
   ]
]);

val bound_child_parent_lemma_1_1_1_3 = store_thm("bound_child_parent_lemma_1_1_1_3",
``!PROB s as vs as'. child_parent_rel(PROB, vs) /\ (s IN (valid_states PROB))
                     /\ as IN (valid_plans PROB) /\ sat_precond_as(s, as) /\ sublist as' (as_proj(as, vs))
                 /\ no_effectless_act(as) /\ sat_precond_as(s, as')
                 ==>
                 (DRESTRICT (exec_plan(s, (replace_projected([], as', as, vs)))) vs
                          = exec_plan(DRESTRICT s vs , as'))``,
SRW_TAC[][]
THEN MP_TAC(bound_child_parent_lemma_1_1_1_2_1
                 |> INST_TYPE [gamma |-> ``:'b``]
                 |> Q.SPEC `PROB`
                 |> Q.SPEC `as`
                 |> Q.SPEC `as'`
                 |> Q.SPEC `vs`)
THEN SRW_TAC[][]
THEN FULL_SIMP_TAC(srw_ss())[]
THEN MP_TAC(bound_child_parent_lemma_1_1_1_3_1
                                |> Q.SPEC `PROB`
                                |> Q.SPEC `vs`
                                |> Q.SPEC `s`
                                |> Q.SPEC `s`
                                |> Q.SPEC `as`
                                |> Q.SPEC `as'`)
THEN SRW_TAC[][]
THEN MP_TAC(graph_plan_lemma_1
                |> Q.SPEC `s`
                |> Q.SPEC `vs`
                |> Q.ISPEC `replace_projected
                              (([] :((α |-> β) # (α |-> β)) list),
                               (as' :((α |-> β) # (α |-> β)) list),
                               (as :((α |-> β) # (α |-> β)) list),(vs :α -> bool))`)
THEN SRW_TAC[][]
THEN METIS_TAC[]);

val bound_child_parent_lemma_1_1_1_4_1_1 = store_thm("bound_child_parent_lemma_1_1_1_4_1_1",
``!PROB vs as_c as. child_parent_rel(PROB, vs) /\ as IN (valid_plans PROB)
                    ==>
                        (FILTER (\a. ~varset_action(a, vs)) as
                                  =   FILTER (\a. ~varset_action(a, vs)) (replace_projected([], as_c, as, vs)))``,
Induct_on `as`
THEN SRW_TAC[][as_proj_def, sublist_def, replace_projected_def]
THEN Cases_on `as_c`
THEN SRW_TAC[][as_proj_def, sublist_def, replace_projected_def]
THEN FULL_SIMP_TAC(bool_ss)[no_effectless_act_def]
THEN SRW_TAC[][FILTER_IDEM]
THENL
[   
   MP_TAC(graph_plan_lemma_22_3_1 |> Q.SPEC `[h]` |> Q.SPEC `[]` |> Q.SPEC `h'::t` |> Q.SPEC `as` |> Q.SPEC `vs`)
   THEN SRW_TAC[][]
   THEN FIRST_X_ASSUM MATCH_MP_TAC
   THEN Q.EXISTS_TAC `PROB`
   THEN SRW_TAC[][]
   THEN FULL_SIMP_TAC(bool_ss)[FDOM_DRESTRICT]
   THEN FULL_SIMP_TAC(bool_ss)[GSYM DISJOINT_DEF]
   THEN METIS_TAC[child_parent_lemma_xx, valid_plan_valid_tail]
   ,
   MP_TAC(graph_plan_lemma_22_3_1 |> Q.SPEC `[h]` |> Q.SPEC `[]` |> Q.SPEC `t` |> Q.SPEC `as` |> Q.SPEC `vs`)
   THEN SRW_TAC[][]
   THEN METIS_TAC[sublist_cons_2, valid_plan_valid_tail]
   ,
   FIRST_X_ASSUM MATCH_MP_TAC
   THEN Q.EXISTS_TAC `PROB`
   THEN SRW_TAC[][]
   THEN FULL_SIMP_TAC(bool_ss)[sublist_def]
   THEN METIS_TAC[valid_plan_valid_tail]
]);

val bound_child_parent_lemma_1_1_1_4_1_2 = store_thm("bound_child_parent_lemma_1_1_1_4_1_2",
``!PROB vs as. child_parent_rel(PROB, vs) /\ no_effectless_act(as) /\ as IN (valid_plans PROB)
                    ==>
                        (FILTER (\a. varset_action(a, (prob_dom PROB) DIFF vs)) as
                                  =   FILTER (\a. ~varset_action(a, vs)) as)``,
NTAC 2 STRIP_TAC
THEN Induct_on `as`
THEN SRW_TAC[][]
THEN1(FIRST_X_ASSUM MATCH_MP_TAC
      THEN SRW_TAC[][]
      THEN FULL_SIMP_TAC(bool_ss)[no_effectless_act_def]
      THEN METIS_TAC[valid_plan_valid_tail])
THEN FULL_SIMP_TAC(bool_ss)[no_effectless_act_def]
THEN METIS_TAC[child_parent_lemma_1_1_2, valid_plan_valid_tail, valid_plan_valid_head]);

val bound_child_parent_lemma_1_1_1_4_1 = store_thm("bound_child_parent_lemma_1_1_1_4_1",
``!PROB vs as_c as. child_parent_rel(PROB, vs) /\ no_effectless_act(as) /\ as IN (valid_plans PROB)
                    ==>
                        (FILTER (\a. varset_action(a, (prob_dom PROB) DIFF vs)) as
                                  =   FILTER (\a. varset_action(a, (prob_dom PROB) DIFF vs)) (replace_projected([], as_c, as, vs)))``,
SRW_TAC[][]
THEN MP_TAC(bound_child_parent_lemma_1_1_1_4_1_1
                   |> Q.SPEC `PROB`
                   |> Q.SPEC `vs`
                   |> Q.SPEC `as_c`
                   |> Q.SPEC `as`)
THEN SRW_TAC[][]
THEN MP_TAC(graph_plan_lemma_22_4
           |> Q.SPEC `PROB`
           |> Q.SPEC `vs`
           |> Q.SPEC `as`
           |> Q.SPEC `as_c`)
THEN SRW_TAC[][]
THEN MP_TAC(bound_child_parent_lemma_1_1_1_4_1_2 |> Q.SPEC `PROB` |> Q.SPEC `vs` |> Q.SPEC `as`)
THEN SRW_TAC[][]
THEN MP_TAC(graph_plan_lemma_22_3
           |> Q.SPEC `PROB`
           |> Q.SPEC `vs`
           |> Q.SPEC `as_c`
           |> Q.SPEC `as`)
THEN SRW_TAC[][]
THEN MP_TAC(graph_plan_lemma_22_4 
            |> Q.SPEC `PROB`
            |> Q.SPEC `vs`
            |> Q.SPEC `as`
            |> Q.SPEC `as_c`)
THEN SRW_TAC[][]
THEN MP_TAC(bound_child_parent_lemma_1_1_1_4_1_2 |> Q.SPEC `PROB` |> Q.SPEC `vs` |> Q.SPEC `(replace_projected ([],as_c,as,vs))`)
THEN SRW_TAC[][]);

val bound_child_parent_lemma_1_1_1_4_2 = store_thm("bound_child_parent_lemma_1_1_1_4_2",
``!PROB vs as. child_parent_rel(PROB, vs) /\ as IN (valid_plans PROB) /\ no_effectless_act(as)
                    ==>
                      (as_proj(as, ((prob_dom PROB) DIFF vs))
                                              = FILTER (\a. varset_action(a, (prob_dom PROB) DIFF vs)) as)``,
SRW_TAC[][]
THEN MP_TAC(child_parent_lemma_2_1_2_1
             |> Q.SPEC `PROB`
             |> Q.SPEC `vs`
             |> Q.SPEC `as`)
THEN SRW_TAC[][]
THEN `(as_proj_parent (as,vs) = as_proj (as, (prob_dom PROB) DIFF vs))` by METIS_TAC[(child_parent_lemma_1_2 
                         |> Q.SPEC `PROB`
                         |> Q.SPEC `as`
                         |> Q.SPEC `vs`)]
THEN FULL_SIMP_TAC(bool_ss)[GSYM FILTER_FILTER, graph_plan_lemma_16_7]
THEN METIS_TAC[((GSYM FILTER_EQ_ID ) |> Q.ISPEC `(λa. FDOM (SND a) ≠ ∅)` |> Q.SPEC `as` |> INST_TYPE [alpha |-> ``:'a|->'b``, beta |-> ``:'a``])]);

val bound_child_parent_lemma_1_1_1_4_3 = store_thm("bound_child_parent_lemma_1_1_1_4_3",
``!PROB vs s as. child_parent_rel(PROB, vs) /\ as IN (valid_plans PROB) /\ (s IN (valid_states PROB)) /\ sat_precond_as(s, as) /\ no_effectless_act(as)
                    ==>
                      sat_precond_as(s, FILTER (\a. varset_action(a, (prob_dom PROB) DIFF vs)) as)``,
SRW_TAC[][]
THEN METIS_TAC[child_parent_lemma_2_1_1_2, child_parent_lemma_2_1_2_1, bound_child_parent_lemma_1_1_1_4_2_1, GSYM FILTER_FILTER]);

val bound_child_parent_lemma_1_1_1_4 = store_thm("bound_child_parent_lemma_1_1_1_4",
``!PROB vs s as_c as. child_parent_rel(PROB, vs) /\ (s IN (valid_states PROB)) /\ as IN (valid_plans PROB) /\ sublist as_c (as_proj(as, vs)) /\ sat_precond_as(s, as)
                    /\ sat_precond_as(s, as_c) /\ no_effectless_act(as)
						      ==>
						                     ((DRESTRICT (exec_plan(s, as)) ((prob_dom PROB) DIFF vs)  
										  = DRESTRICT (exec_plan(s, (replace_projected([], as_c, as, vs)))) 
											        ((prob_dom PROB) DIFF vs))
                        /\ (FILTER (\a. varset_action (a,(prob_dom PROB) DIFF vs)) as =
                                      FILTER (\a. varset_action (a,(prob_dom PROB) DIFF vs))
                                                   (replace_projected ([],as_c,as,vs))))``,
SRW_TAC[][]
THEN MP_TAC(GSYM (bound_child_parent_lemma_1_1_1_4_1
                  |> INST_TYPE[gamma |-> ``:'b``]
                  |> Q.SPEC `PROB`
                  |> Q.SPEC `vs`
                  |> Q.SPEC `as_c`
                  |> Q.SPEC `as`))
THEN SRW_TAC[][]
THEN MP_TAC(GSYM (bound_child_parent_lemma_1_1_1_4_2
                  |> INST_TYPE[gamma |-> ``:'b``]
                  |> Q.SPEC `PROB`
                  |> Q.SPEC `vs`
                  |> Q.SPEC `as`))
THEN SRW_TAC[][]
THEN FULL_SIMP_TAC(bool_ss)[]
THEN MP_TAC(GSYM (bound_child_parent_lemma_1_1_1_4_3
                  |> Q.SPEC `PROB`
                  |> Q.SPEC `vs`
                  |> Q.SPEC `s`
                  |> Q.SPEC `as`))
THEN SRW_TAC[][]
THEN MP_TAC(graph_plan_lemma_11
           |> Q.SPEC `s`
           |> Q.SPEC `as`
           |> Q.SPEC `((prob_dom (PROB:((α |-> β) # (α |-> β)) set)) DIFF vs)`)
THEN SRW_TAC[][]
THEN MP_TAC(bound_child_parent_lemma_1_1_1_3_1
                                |> Q.SPEC `PROB`
                                |> Q.SPEC `vs`
                                |> Q.SPEC `s`
                                |> Q.SPEC `s`
                                |> Q.SPEC `as`
                                |> Q.SPEC `as_c`)
THEN SRW_TAC[][]
THEN MP_TAC(graph_plan_lemma_11
                |> Q.SPEC `s`
                |> Q.ISPEC `replace_projected ([]:((α |-> β) # (α |-> β)) list,as_c,as,vs)`
                |> Q.SPEC `((prob_dom (PROB:((α |-> β) # (α |-> β)) set)) DIFF vs)`)
THEN SRW_TAC[][]
THEN MP_TAC(graph_plan_lemma_22_4
           |> INST_TYPE[gamma |-> ``:'b``]
           |> Q.SPEC `PROB`
           |> Q.SPEC `vs`
           |> Q.SPEC `as`
           |> Q.SPEC `as_c`)
THEN SRW_TAC[][]
THEN MP_TAC(graph_plan_lemma_22_3 
            |> INST_TYPE[gamma |-> ``:'b``]
            |> Q.SPEC `PROB`
            |> Q.SPEC `vs`
            |> Q.SPEC `as_c`
            |> Q.SPEC `as`)
THEN SRW_TAC[][]
THEN MP_TAC(GSYM(bound_child_parent_lemma_1_1_1_4_2
                  |> INST_TYPE[gamma |-> ``:'b``]
                  |> Q.SPEC `PROB`
                  |> Q.SPEC `vs`
                  |> Q.ISPEC `replace_projected ([]:((α |-> β) # (α |-> β)) list,as_c,as,vs)`))
THEN SRW_TAC[][]
THEN FULL_SIMP_TAC(bool_ss)[]);

(*val bound_child_parent_lemma_1_1_1 = store_thm("bound_child_parent_lemma_1_1_1",
``!PROB as vs s. FINITE PROB /\ (s IN (valid_states PROB)) /\ as IN (valid_plans PROB) /\ child_parent_rel(PROB, vs)
 ==>  ?as'. (exec_plan(s, as') = exec_plan(s, as))
          /\ LENGTH (FILTER (\a. varset_action(a, vs)) as') <= problem_plan_bound(prob_proj(PROB, vs))
    /\ (sublist as' as)``,
SRW_TAC[][]
THEN `rem_effectless_act (rem_condless_act (s,[],as)) IN valid_plans PROB`
       by METIS_TAC[graph_plan_lemma_7_10, graph_plan_lemma_16_4']
THEN MP_TAC (bound_child_parent_lemma_1_1_1_1
|> Q.SPEC `PROB`
|> Q.SPEC `s`
|> Q.SPEC `(rem_effectless_act(rem_condless_act(s, [], as)))`
|> Q.SPEC `vs`)
THEN SRW_TAC[][graph_plan_lemma_7_4, graph_plan_lemma_16_4, SUBSET_TRANS, graph_plan_lemma_7_2, graph_plan_lemma_16_2]
THEN Q.EXISTS_TAC `(replace_projected ([],as',(rem_effectless_act(rem_condless_act(s, [], as))), vs))`
THEN SRW_TAC[][]
THENL
[
ASSUME_TAC (MATCH_MP (graph_plan_lemma_16_2 |> Q.SPEC `rem_condless_act(s, [], as)` |> Q.SPEC `s`) (graph_plan_lemma_7_2 |> Q.SPEC `(as)` |> Q.SPEC `s`))
THEN MP_TAC(bound_child_parent_lemma_1_1_1_3
         |> Q.SPEC `PROB`
 |> Q.SPEC `s`
 |> Q.SPEC `(rem_effectless_act(rem_condless_act(s, [], as)))`
 |> Q.SPEC `vs`
 |> Q.SPEC `as'`)
THEN SRW_TAC[][GSYM valid_plans_def, graph_plan_lemma_7_4, graph_plan_lemma_16_4, SUBSET_TRANS, graph_plan_lemma_16_6]
THEN MP_TAC(graph_plan_lemma_1 
           |> Q.SPEC `s` 
      |> Q.SPEC `vs` 
      |> Q.SPEC `(rem_effectless_act(rem_condless_act(s, [], as)))`)
THEN SRW_TAC[][]
THEN FULL_SIMP_TAC(bool_ss)[]
THEN MP_TAC(bound_child_parent_lemma_1_1_1_4
         |> Q.SPEC `PROB`
 |> Q.SPEC `vs`
 |> Q.SPEC `s`
 |> Q.SPEC `as'`
 |> Q.SPEC `(rem_effectless_act(rem_condless_act(s, [], as)))`)
THEN SRW_TAC[][graph_plan_lemma_7_4, graph_plan_lemma_16_4, SUBSET_TRANS, graph_plan_lemma_16_6]
THEN MP_TAC(graph_plan_lemma_6_1
|> Q.SPEC `(rem_effectless_act(rem_condless_act(s, [], as)))`
|> Q.SPEC `s`
|> Q.SPEC `PROB`)
THEN SRW_TAC[][graph_plan_lemma_7_4, graph_plan_lemma_16_4, SUBSET_TRANS]
THEN MP_TAC (graph_plan_lemma_22_1'
|> Q.SPEC `(rem_effectless_act(rem_condless_act(s, [], as)))`
|> INST_TYPE [beta |-> ``:'a state``]
|> Q.SPEC `as'`
|> Q.SPEC `[]`
|> Q.SPEC `vs`
|> Q.SPEC `PROB`)
THEN SRW_TAC[][graph_plan_lemma_7_4, graph_plan_lemma_16_4, SUBSET_TRANS, empty_plan_is_valid]
THEN MP_TAC (graph_plan_lemma_6_1
|> Q.SPEC `replace_projected ([],as', (rem_effectless_act(rem_condless_act(s, [], as))), vs)`
|> Q.SPEC `s`
|> Q.SPEC `PROB`)
THEN SRW_TAC[][graph_plan_lemma_7_4, graph_plan_lemma_16_4, SUBSET_TRANS]
        THEN FULL_SIMP_TAC(srw_ss())[valid_states_def]
THEN MP_TAC(graph_plan_lemma_5
|> Q.ISPEC `(exec_plan (s, (replace_projected ([],as',(rem_effectless_act(rem_condless_act(s, [], as))), vs))))`
|> Q.ISPEC `(exec_plan (s, (rem_effectless_act(rem_condless_act(s, [], as)))))`
|> Q.SPEC `vs`)
THEN SRW_TAC[][]
THEN SRW_TAC[][GSYM graph_plan_lemma_7_1, GSYM graph_plan_lemma_16_1]
,
MP_TAC(bound_child_parent_lemma_1_1_1_2
     |> INST_TYPE [beta |-> ``:bool``]
     |> Q.SPEC `PROB`
     |> Q.SPEC `s`
     |> Q.SPEC `(rem_effectless_act(rem_condless_act(s, [], as)))`
     |> Q.SPEC `vs`
     |> Q.SPEC `as'`)
        THEN SRW_TAC[][graph_plan_lemma_7_4, graph_plan_lemma_16_4, SUBSET_TRANS, graph_plan_lemma_16_6]
        THEN FULL_SIMP_TAC(bool_ss)[]
,
METIS_TAC[graph_plan_lemma_7_8, graph_plan_lemma_16_9, graph_plan_lemma_22_2, sublist_trans]
]);
*)

val bound_child_parent_lemma_1_1_1' = store_thm("bound_child_parent_lemma_1_1_1'",
``!(PROB:'a problem) as vs s. FINITE PROB /\ (s IN (valid_states PROB)) /\ as IN (valid_plans PROB)
									    /\ child_parent_rel(PROB, vs) /\ sat_precond_as(s, as)
                 /\ no_effectless_act(as)
		     ==>  ?as'. (exec_plan(s, as') = exec_plan(s, as))
		         /\ (LENGTH (FILTER (\a. varset_action(a, vs)) as') <= problem_plan_bound(prob_proj(PROB, vs))
                                /\ ((FILTER (\a. varset_action(a, (prob_dom PROB) DIFF vs)) as')
                                             = (FILTER (\a. varset_action(a, (prob_dom PROB) DIFF vs)) as)))
			        /\ (sublist as' as) /\ no_effectless_act(as') /\ as' IN (valid_plans PROB) /\ sat_precond_as(s, as')``,
SRW_TAC[][]
THEN `rem_effectless_act (rem_condless_act (s,[],as)) IN valid_plans PROB`
       by METIS_TAC[graph_plan_lemma_7_10, graph_plan_lemma_16_4']
THEN MP_TAC (bound_child_parent_lemma_1_1_1_1
		 |> Q.SPEC `PROB`
		 |> Q.SPEC `s`
		 |> Q.SPEC `as`
		 |> Q.SPEC `vs`)
THEN SRW_TAC[][graph_plan_lemma_7_4, graph_plan_lemma_16_4, SUBSET_TRANS, graph_plan_lemma_7_2, graph_plan_lemma_16_2]
THEN Q.EXISTS_TAC `(replace_projected ([],as',as, vs))`
THEN SRW_TAC[][]
THENL
[
  ASSUME_TAC((MATCH_MP (graph_plan_lemma_16_2|> Q.SPEC `rem_condless_act(s, [], as)` |> Q.SPEC `s`) (graph_plan_lemma_7_2 |> Q.SPEC `(as)` |> Q.SPEC `s`))
                   |> INST_TYPE[beta |-> ``:bool``])
	    THEN MP_TAC(bound_child_parent_lemma_1_1_1_3
                                 |> INST_TYPE[beta |-> ``:bool``]
				          |> Q.SPEC `PROB`
					   |> Q.SPEC `s`
					    |> Q.SPEC `as`
					     |> Q.SPEC `vs`
					      |> Q.SPEC `as'`)
	    THEN SRW_TAC[][graph_plan_lemma_7_4, graph_plan_lemma_16_4, SUBSET_TRANS, graph_plan_lemma_16_6]
	    THEN MP_TAC(graph_plan_lemma_1 
                              |> INST_TYPE[beta |-> ``:bool``]
			            |> Q.SPEC `s` 
				          |> Q.SPEC `vs` 
					        |> Q.SPEC `as`)
	    THEN SRW_TAC[][]
	    THEN FULL_SIMP_TAC(bool_ss)[]
	    THEN MP_TAC(bound_child_parent_lemma_1_1_1_4
                                 |> INST_TYPE[beta |-> ``:bool``]
				          |> Q.SPEC `PROB`
					   |> Q.SPEC `vs`
					    |> Q.SPEC `s`
					     |> Q.SPEC `as'`
					      |> Q.SPEC `as`)
	    THEN SRW_TAC[][graph_plan_lemma_7_4, graph_plan_lemma_16_4, SUBSET_TRANS, graph_plan_lemma_16_6]
	    THEN MP_TAC(graph_plan_lemma_6_1
                        |> INST_TYPE[beta |-> ``:bool``]
			|> Q.SPEC `as`
			|> Q.SPEC `s`
			|> Q.SPEC `PROB`)
	    THEN SRW_TAC[][graph_plan_lemma_7_4, graph_plan_lemma_16_4, SUBSET_TRANS]
	    THEN MP_TAC (graph_plan_lemma_22_1'
			     |> Q.SPEC `as`
			     |> INST_TYPE [beta |-> ``:bool``, gamma |-> ``:bool``]
			     |> Q.SPEC `as'`
			     |> Q.SPEC `[]`
			     |> Q.SPEC `vs`
			     |> Q.SPEC `PROB`)
	    THEN SRW_TAC[][graph_plan_lemma_7_4, graph_plan_lemma_16_4, SUBSET_TRANS, empty_plan_is_valid]
	    THEN MP_TAC (graph_plan_lemma_6_1
                                |> INST_TYPE[beta |-> ``:bool``]
				|> Q.SPEC `replace_projected ([],as', as, vs)`
				|> Q.SPEC `s`
				|> Q.SPEC `PROB`)
	    THEN SRW_TAC[][graph_plan_lemma_7_4, graph_plan_lemma_16_4, SUBSET_TRANS]
        THEN FULL_SIMP_TAC(srw_ss())[valid_states_def]
	THEN MP_TAC(graph_plan_lemma_5
                                |> INST_TYPE[beta |-> ``:bool``]
				|> Q.SPEC `(exec_plan (s:'a state, (replace_projected ([]:'a action list,as',as, vs))))`
				|> Q.SPEC `(exec_plan (s, as))`
				|> Q.SPEC `vs`)
	THEN SRW_TAC[][]
	,
	  MP_TAC(bound_child_parent_lemma_1_1_1_2
		          |> INST_TYPE [beta |-> ``:bool``, gamma |-> ``:bool``, delta |-> ``:bool``]
			       |> Q.SPEC `PROB`
			            |> Q.SPEC `s`
				         |> Q.SPEC `as`
					      |> Q.SPEC `vs`
					           |> Q.SPEC `as'`)
        THEN SRW_TAC[][graph_plan_lemma_7_4, graph_plan_lemma_16_4, SUBSET_TRANS, graph_plan_lemma_16_6]
        THEN FULL_SIMP_TAC(bool_ss)[]
	,
        MP_TAC(bound_child_parent_lemma_1_1_1_4_1
		            |> INST_TYPE [beta |-> ``:bool``, gamma |-> ``:bool``]
			             |> Q.SPEC `PROB`
				      |> Q.SPEC `vs`
				       |> Q.SPEC `as'`
				        |> Q.SPEC `as`)
	      THEN SRW_TAC[][graph_plan_lemma_7_4, graph_plan_lemma_16_4, SUBSET_TRANS, graph_plan_lemma_16_6]
        ,
        METIS_TAC[graph_plan_lemma_7_8, graph_plan_lemma_16_9, graph_plan_lemma_22_2, sublist_trans]
        ,
        METIS_TAC[graph_plan_lemma_22_3]
        ,
        METIS_TAC[graph_plan_lemma_22_4]
        ,
        METIS_TAC[bound_child_parent_lemma_1_1_1_3_1]
]);

val parent_children_lemma_1_1_1_11_5 = store_thm("parent_children_lemma_1_1_1_11_5",
`` ∀P P2 P3 vs. (!a. P3 a /\ P2 (action_proj(a,vs)) ==> (P a <=> P (action_proj(a,vs))))
       /\ (!a. P a /\ P2 a /\ P3 a ==> varset_action(a, vs))
       /\ (!a. P2 a /\ P3 a /\ ~varset_action(a, vs) ==> (DISJOINT (FDOM (SND a)) vs))
        ==> !as' as''. sublist (as') (as_proj(as'', vs)) /\ EVERY P2 as' /\ EVERY P2 as'' /\ EVERY P3 as''
                               ==> (LENGTH (FILTER P (replace_projected([],as',as'',vs)))
                                      = LENGTH (FILTER P as'))``,
NTAC 5 STRIP_TAC
THEN Induct_on `as''`
THEN1 (Cases_on `as'`
      THEN SRW_TAC[][replace_projected_def, as_proj_def, sublist_def])
THEN1(SRW_TAC[][replace_projected_def]
      THEN Cases_on `as'`
      THEN SRW_TAC[][replace_projected_def]
      THEN1(MP_TAC(SIMP_RULE(srw_ss())[EQ_IMP_THM] (graph_plan_lemma_21 |> INST_TYPE [alpha |-> ``:('a |-> 'b) # ('a |-> 'c)``]|> Q.SPECL[`P2`,`P3`, `as''`]))
            THEN SRW_TAC[][]
            THEN METIS_TAC[(((parent_children_lemma_1_1_1_11_5_2
                               |> Q.SPEC `P`)
                               |> Q.SPEC `(\a. P2 a /\ P3 a)`)
                               |> Q.ISPEC `(λa. varset_action (a,vs))`)
                           , LENGTH_0])
      THEN1(MP_TAC(SIMP_RULE(srw_ss())[EQ_IMP_THM] (graph_plan_lemma_21 |> INST_TYPE [alpha |-> ``:('a |-> 'b) # ('a |-> 'c)``]|> Q.SPECL[`P2`,`P3`, `as''`]))
            THEN SRW_TAC[][]
            THEN METIS_TAC[(((parent_children_lemma_1_1_1_11_5_2
                               |> Q.SPEC `P`)
                               |> Q.SPEC `(\a. P2 a /\ P3 a)`)
                               |> Q.ISPEC `(λa. varset_action (a,vs))`)
                           , LENGTH_0])
      THEN1(REWRITE_TAC[REWRITE_RULE[APPEND_NIL](graph_plan_lemma_22_3_1
						     |> Q.SPEC `[h]`
						     |> Q.SPEC `[]`
						     |> Q.SPEC `t`
						     |> Q.SPEC `as''`
						     |> Q.SPEC `vs`)]
            THEN SRW_TAC[][]
            THEN SRW_TAC[][FILTER_APPEND_DISTRIB]
            THEN1(FIRST_X_ASSUM MATCH_MP_TAC
                  THEN SRW_TAC[][]
                  THEN1(FULL_SIMP_TAC(srw_ss())[as_proj_def]
                        THEN METIS_TAC[sublist_def, sublist_cons_2, sublist_cons_3])
                  THEN1 FULL_SIMP_TAC(srw_ss())[]  )
            THEN1(FULL_SIMP_TAC(srw_ss())[]
                  THEN METIS_TAC[]))
      THEN1(REWRITE_TAC[REWRITE_RULE[APPEND_NIL](graph_plan_lemma_22_3_1
						     |> Q.SPEC `[h]`
						     |> Q.SPEC `[]`
						     |> Q.SPEC `t`
						     |> Q.SPEC `as''`
						     |> Q.SPEC `vs`)]
            THEN SRW_TAC[][]
            THEN SRW_TAC[][FILTER_APPEND_DISTRIB]
            THEN1(FULL_SIMP_TAC(srw_ss())[]
                  THEN METIS_TAC[])
            THEN1(FIRST_X_ASSUM MATCH_MP_TAC
                  THEN SRW_TAC[][]
                  THEN FULL_SIMP_TAC(srw_ss())[as_proj_def]
                  THEN METIS_TAC[sublist_def, sublist_cons_2, sublist_cons_3]))
      THEN1(FIRST_X_ASSUM (MP_TAC o Q.SPECL[`h'::t`])
            THEN SRW_TAC[][] 
            THEN FIRST_X_ASSUM MATCH_MP_TAC
            THEN FULL_SIMP_TAC(srw_ss())[as_proj_def]
            THEN METIS_TAC[sublist_def, sublist_cons_2, sublist_cons_3])
      THEN1(FIRST_X_ASSUM (MP_TAC o Q.SPECL[`h'::t`])
            THEN SRW_TAC[][]
            THEN FIRST_X_ASSUM MATCH_MP_TAC
            THEN FULL_SIMP_TAC(srw_ss())[as_proj_def]
            THEN METIS_TAC[sublist_def, sublist_cons_2, sublist_cons_3])
      THEN1(REWRITE_TAC[REWRITE_RULE[APPEND_NIL](graph_plan_lemma_22_3_1
						     |> Q.SPEC `[h]`
						     |> Q.SPEC `[]`
						     |> Q.SPEC `t`
						     |> Q.SPEC `as''`
						     |> Q.SPEC `vs`)]
            THEN SRW_TAC[][]
            THEN SRW_TAC[][FILTER_APPEND_DISTRIB]
            THEN1 METIS_TAC[]
            THEN1(FIRST_X_ASSUM (MP_TAC o Q.SPECL[`h'::t`])
                  THEN SRW_TAC[][]
                  THEN FIRST_X_ASSUM MATCH_MP_TAC
                  THEN FULL_SIMP_TAC(srw_ss())[as_proj_def]
                  THEN METIS_TAC[sublist_def, sublist_cons_2, sublist_cons_3])) 
      THEN1(REWRITE_TAC[REWRITE_RULE[APPEND_NIL](graph_plan_lemma_22_3_1
						     |> Q.SPEC `[h]`
						     |> Q.SPEC `[]`
						     |> Q.SPEC `t`
						     |> Q.SPEC `as''`
						     |> Q.SPEC `vs`)]
            THEN SRW_TAC[][]
            THEN SRW_TAC[][FILTER_APPEND_DISTRIB]
            THEN1(METIS_TAC[])
            THEN1(FIRST_X_ASSUM (MP_TAC o Q.SPECL[`h'::t`])
                  THEN SRW_TAC[][]
                  THEN FIRST_X_ASSUM MATCH_MP_TAC
                  THEN FULL_SIMP_TAC(srw_ss())[as_proj_def, FDOM_DRESTRICT, DISJOINT_DEF]
                  THEN MP_TAC(sublist_every |> INST_TYPE [alpha |-> ``:'a state # 'a state``] |> Q.SPECL [`h::as''`, `as'''`, `P2`])
                  THEN SRW_TAC[][]
                  THEN METIS_TAC[sublist_def, sublist_cons_2, sublist_cons_3]))))


val bound_child_parent_lemma_1_1_1'' = store_thm("bound_child_parent_lemma_1_1_1''",
``!PROB as as' vs s. child_parent_rel(PROB, vs) /\ s IN valid_states PROB
                     /\ no_effectless_act(as) /\ sublist as' (as_proj(as, vs)) 
                     /\ as IN valid_plans PROB /\ (DRESTRICT (exec_plan(s,as)) vs = (exec_plan(DRESTRICT s vs,as')))
                     /\ sat_precond_as(s, as) /\ sat_precond_as(s, as')
                    ==>  (exec_plan(s, replace_projected ([],as',as,vs)) = exec_plan(s, as))``,
SRW_TAC[][]
THEN MP_TAC(bound_child_parent_lemma_1_1_1_3 |> Q.SPECL[`PROB`, `s`, `as`, `vs`, `as'`])
THEN SRW_TAC[][]
THEN MP_TAC(bound_child_parent_lemma_1_1_1_4 |> Q.SPECL[`PROB`, `vs`, `s`,`as'`, `as`])
THEN SRW_TAC[][]
THEN MP_TAC(graph_plan_lemma_6_1 |> Q.SPECL[`as`, `s`, `PROB`])
THEN SRW_TAC[][]
THEN MATCH_MP_TAC(graph_plan_lemma_5)
THEN Q.EXISTS_TAC `vs`
THEN SRW_TAC[][]
THEN `(exec_plan (s,replace_projected ([],as',as,vs))) IN valid_states PROB` by METIS_TAC[graph_plan_lemma_22_2, sublist_valid_plan, graph_plan_lemma_6_1]
THEN FULL_SIMP_TAC(srw_ss())[valid_states_def])

val _ = export_theory();
