open HolKernel Parse boolLib bossLib;

open finite_mapTheory
open arithmeticTheory
open pred_setTheory
open rich_listTheory
open listTheory;							 
open utilsTheory;
open factoredSystemTheory 
open sublistTheory;
open parentChildStructureTheory;
open systemAbstractionTheory;
(* open depGraphVtxCutTheory; *)
open factoredSystemTheory;
open topologicalPropsTheory;
open prim_recTheory;
open dependencyTheory
open set_utilsTheory

val _ = new_theory "parentTwoChildrenStructure";

val two_children_parent_rel_def = 
      Define`two_children_parent_rel(PROB, vs_1, vs_2)
                <=> ~(dep_var_set(PROB, vs_1, (prob_dom PROB) DIFF vs_1)) /\ ~(dep_var_set(PROB, vs_2, (prob_dom PROB) DIFF vs_2))`;
val two_children_parent_bound_main_lemma_1 = store_thm("two_children_parent_bound_main_lemma_1",
``!PROB vs_1 vs_2. two_children_parent_rel(PROB, vs_1, vs_2)
                   ==> child_parent_rel(PROB, vs_1 UNION vs_2)``,
SRW_TAC[][two_children_parent_rel_def, child_parent_rel_def, dep_var_set_def, DISJOINT_DEF, UNION_DEF,
          INTER_DEF, DIFF_DEF, GSPEC_ETA, EMPTY_DEF]
THEN METIS_TAC[])

val indep_rel_def = 
      Define`indep_rel(PROB, vs_1)
                <=> ~(dep_var_set(PROB, vs_1, (prob_dom PROB) DIFF vs_1)) /\ ~(dep_var_set(PROB, (prob_dom PROB) DIFF vs_1, vs_1))`;

val two_children_parent_bound_main_lemma_2 = store_thm("two_children_parent_bound_main_lemma_2",
``!PROB vs_1 vs_2. vs_1 SUBSET (prob_dom PROB) /\ vs_2 SUBSET (prob_dom PROB) /\ DISJOINT vs_1 vs_2
                   /\ two_children_parent_rel(PROB, vs_1, vs_2)
                   ==> indep_rel(prob_proj(PROB, vs_1 UNION vs_2), vs_1)``,
SRW_TAC[][two_children_parent_rel_def, indep_rel_def]
THEN `(vs_1 UNION vs_2) SUBSET (prob_dom PROB)` by SRW_TAC[][]
THEN MP_TAC(two_children_parent_bound_main_lemma_2_1 
               |> Q.SPECL[`PROB`, `vs_1 UNION vs_2`])
THEN SRW_TAC[][]
THEN FULL_SIMP_TAC(srw_ss())[]
THEN1( `(vs_1 UNION vs_2) DIFF vs_1 = vs_2` 
           by (SRW_TAC[][UNION_DEF, DIFF_DEF, EXTENSION, IN_DEF] 
               THEN FULL_SIMP_TAC(srw_ss())[DISJOINT_DEF, INTER_DEF, EXTENSION, IN_DEF] 
               THEN METIS_TAC[])
       THEN SRW_TAC[][]
       THEN `vs_2 SUBSET ((prob_dom PROB) DIFF vs_1)` by 
             (FULL_SIMP_TAC(srw_ss())[SUBSET_DEF, DIFF_DEF, DISJOINT_DEF, EXTENSION]
              THEN METIS_TAC[])
       THEN METIS_TAC[two_children_parent_bound_main_lemma_2_2, scc_lemma_1_4_1_1_2_2_1,
                      two_children_parent_bound_main_lemma_2_3, DISJOINT_SYM])
THEN1( `(vs_1 UNION vs_2) DIFF vs_1 = vs_2` 
           by (SRW_TAC[][UNION_DEF, DIFF_DEF, EXTENSION, IN_DEF] 
               THEN FULL_SIMP_TAC(srw_ss())[DISJOINT_DEF, INTER_DEF, EXTENSION, IN_DEF] 
               THEN METIS_TAC[])
       THEN SRW_TAC[][]
       THEN `vs_1 SUBSET ((prob_dom PROB) DIFF vs_2)` by 
             (FULL_SIMP_TAC(srw_ss())[SUBSET_DEF, DIFF_DEF, DISJOINT_DEF, EXTENSION]
              THEN METIS_TAC[])
       THEN METIS_TAC[two_children_parent_bound_main_lemma_2_2, scc_lemma_1_4_1_1_2_2_1,
                      two_children_parent_bound_main_lemma_2_3, DISJOINT_SYM]))

val two_children_parent_bound_main_lemma_3_1_1 = store_thm("two_children_parent_bound_main_lemma_3_1_1",
``!PROB vs. vs SUBSET (prob_dom PROB) /\ indep_rel(PROB, vs)
            ==> child_parent_rel(PROB, vs) /\ child_parent_rel(PROB, (prob_dom PROB) DIFF vs)``,
SRW_TAC[][indep_rel_def, child_parent_rel_def]
THEN METIS_TAC[two_children_parent_bound_main_lemma_3_1_2])

(*val two_children_parent_bound_main_lemma_3_1 = store_thm("two_children_parent_bound_main_lemma_3_1",
``!PROB as. indep_rel(PROB,vs) /\ plan(PROB, as) /\ FINITE vs /\ vs SUBSET (prob_dom PROB)
            ==> ?as'. plan(PROB, as') /\ sublist as' as
                      /\ LENGTH as' 
                            <= problem_plan_bound(prob_proj(PROB, vs))
                               + problem_plan_bound(prob_proj(PROB, (prob_dom PROB) DIFF vs))``,
SRW_TAC[][plan_def]
THEN MP_TAC(two_children_parent_bound_main_lemma_3_1_1 |> Q.SPECL[`PROB`, `vs`])
THEN SRW_TAC[][]
THEN `set (rem_condless_act(PROB.I, [], as)) SUBSET PROB.A` by METIS_TAC[graph_plan_lemma_7_4]
THEN `sat_precond_as (PROB.I, rem_condless_act(PROB.I, [], as))` by METIS_TAC[graph_plan_lemma_7_2]
THEN MP_TAC(graph_plan_lemma_16_2|> Q.SPECL[`rem_condless_act(PROB.I, [], as)`,`PROB.I`] )
THEN SRW_TAC[][]
THEN `set(rem_effectless_act (rem_condless_act (PROB.I,[],as))) 
           SUBSET PROB.A` by METIS_TAC[graph_plan_lemma_16_4]
THEN `no_effectless_act(rem_effectless_act (rem_condless_act (PROB.I,[],as)))` by 
           METIS_TAC[graph_plan_lemma_16_6]
THEN MP_TAC(bound_child_parent_lemma_1_1_1' 
               |> Q.SPECL[`PROB`, 
                          `(rem_effectless_act (rem_condless_act (PROB.I,[],as)))`,
                          `vs`, `PROB.I`])
THEN SRW_TAC[][]
THEN MP_TAC(bound_child_parent_lemma_1_1_1' 
               |> Q.SPECL[`PROB`, 
                          `as'`,
                          `FDOM PROB.I DIFF vs`, `PROB.I`])
THEN SRW_TAC[][]
THEN Q.EXISTS_TAC `as''`
THEN SRW_TAC[][]
THEN1(METIS_TAC[graph_plan_lemma_16_1, graph_plan_lemma_7_1])
THEN1(METIS_TAC[sublist_trans, graph_plan_lemma_16_9, graph_plan_lemma_7_8])
THEN1(
      MP_TAC(two_children_parent_bound_main_lemma_3_1_2 |> Q.SPECL[`FDOM PROB.I`, `vs`])
      THEN SRW_TAC[][]
      THEN FULL_SIMP_TAC(srw_ss())[]
      THEN MP_TAC(bound_child_parent_lemma_1_1_1_4_1_2 |> Q.SPECL[`PROB`, `vs`, `as''`])
      THEN SRW_TAC[][]
      THEN FULL_SIMP_TAC(bool_ss)[]
      THEN MP_TAC(SIMP_RULE(srw_ss())[](two_children_parent_bound_main_lemma_3_1_3 |> Q.ISPECL[`(λa. varset_action (a,vs))`]) 
       |> Q.SPEC `as''`)
      THEN SRW_TAC[][]
      THEN DECIDE_TAC))

val two_children_parent_bound_main_lemma_3 = store_thm("two_children_parent_bound_main_lemma_3",
``!vs. FINITE vs 
        ==> (!PROB. planning_problem(PROB) /\ indep_rel(PROB, vs) /\ vs SUBSET FDOM PROB.I
                   ==> (problem_plan_bound PROB <=
                                ((problem_plan_bound 
                                        (prob_proj 
                                            (PROB, vs))) 
                                   + (problem_plan_bound 
                                        (prob_proj 
                                            (PROB, FDOM PROB.I DIFF vs))))))``,
NTAC 2 STRIP_TAC
THEN REWRITE_TAC[LE_LT1]
THEN HO_MATCH_MP_TAC bound_child_parent_lemma_2
THEN SRW_TAC[][]
THEN1 FULL_SIMP_TAC(srw_ss())[indep_rel_def, dep_var_set_def, dep_def]
THEN1(SRW_TAC[][problem_plan_bound_def, MPLS_def, PLS_def, prob_proj_def]
      THEN MP_TAC(bound_child_parent_lemma_1_1_1_1_3 |> INST_TYPE [alpha |-> ``:'a``, beta |-> ``:bool``, gamma |-> ``:bool``] |> Q.SPEC `s` |> Q.SPEC `PROB.I`)
      THEN SRW_TAC[][])
THEN1 METIS_TAC[two_children_parent_bound_main_lemma_3_1, LE_LT1])

val two_children_parent_bound_main_lemma = store_thm("two_children_parent_bound_main_lemma",
``!PROB vs_1 vs_2. FINITE vs_1 /\ FINITE vs_2 /\ planning_problem(PROB) /\ DISJOINT vs_1 vs_2
                   /\ vs_1 SUBSET FDOM PROB.I /\ vs_2 SUBSET FDOM PROB.I
                   /\ PROB.A SUBSET (\a. ~(FDOM (SND a) = EMPTY)) /\ two_children_parent_rel(PROB, vs_1, vs_2)
                   ==> problem_plan_bound PROB <
                                (problem_plan_bound 
                                        (prob_proj 
                                            (PROB,FDOM PROB.I DIFF (vs_1 UNION vs_2)))
                                             + 1)
                                * ((problem_plan_bound 
                                        (prob_proj 
                                            (PROB, vs_1))) 
                                   + (problem_plan_bound 
                                        (prob_proj 
                                            (PROB, vs_2)))
                                   + 1)``,
SRW_TAC[][]
THEN MP_TAC(two_children_parent_bound_main_lemma_1 |> Q.SPECL[`PROB`, `vs_1`, `vs_2`])
THEN SRW_TAC[][]
THEN MP_TAC(bound_child_parent_lemma |> Q.SPECL[`PROB`, `vs_1 UNION vs_2`])
THEN SRW_TAC[][]
THEN MP_TAC(two_children_parent_bound_main_lemma_2 |> Q.SPECL[`PROB`,`vs_1`,`vs_2`])
THEN SRW_TAC[][]
THEN MP_TAC(two_children_parent_bound_main_lemma_3 |> Q.SPECL[`vs_1`])
THEN SRW_TAC[][]
THEN MP_TAC( graph_plan_lemma_2_2 |> Q.SPECL[`PROB`, `vs_1 UNION vs_2`])
THEN SRW_TAC[][]
THEN `(vs_1 UNION vs_2) SUBSET FDOM PROB.I` by SRW_TAC[][]
THEN MP_TAC(two_children_parent_bound_main_lemma_2_1 |> Q.SPECL[`PROB`, `vs_1 UNION vs_2`])
THEN SRW_TAC[][]
THEN Q.PAT_X_ASSUM `!x. P` (MP_TAC o Q.SPEC `prob_proj(PROB, vs_1 UNION vs_2)`)
THEN SRW_TAC[][]
THEN FULL_SIMP_TAC(bool_ss)[two_children_parent_bound_main_lemma_4]
THEN `vs_1 UNION vs_2 SUBSET FDOM PROB.I` by SRW_TAC[][]
THEN FULL_SIMP_TAC(bool_ss)[two_children_parent_bound_main_lemma_2_1]
THEN MP_TAC( REWRITE_RULE[Once UNION_COMM]  (MATCH_MP (AND2_THM) (DIFF_SAME_UNION |> Q.SPECL[`vs_2`,`vs_1`])) )
THEN SRW_TAC[][]
THEN FULL_SIMP_TAC(bool_ss)[]
THEN MP_TAC(two_children_parent_bound_main_lemma_6 |> Q.SPECL[`vs_2`, `vs_1`])
THEN SRW_TAC[][DISJOINT_SYM]
THEN FULL_SIMP_TAC(bool_ss)[]
THEN `(prob_proj (prob_proj (PROB,vs_1 ∪ vs_2),vs_2)) = 
             (prob_proj (PROB, vs_2))` 
         by METIS_TAC[two_children_parent_bound_main_lemma_4,
               UNION_COMM]
THEN FULL_SIMP_TAC(srw_ss())[]
THEN REWRITE_TAC[Once MULT_COMM]
THEN FULL_SIMP_TAC(srw_ss()) [two_children_parent_bound_main_lemma_8]
THEN MP_TAC(two_children_parent_bound_main_lemma_9
                |> Q.SPECL[`problem_plan_bound (prob_proj (PROB,FDOM PROB.I DIFF (vs_1 ∪ vs_2)))`,
                           `(problem_plan_bound (prob_proj (PROB,vs_1)) + problem_plan_bound (prob_proj (PROB,vs_2)))`,
                           `problem_plan_bound (prob_proj (PROB,vs_1 ∪ vs_2))`])
THEN ASM_SIMP_TAC(bool_ss)[]
THEN STRIP_TAC
THEN DECIDE_TAC);*)

val _ = export_theory();